import time
from typing import TYPE_CHECKING, Generator, Optional, Union

import ingenialogger
from ingenialink.exceptions import ILError

if TYPE_CHECKING:
    from ingeniamotion.motion_controller import MotionController
from ingeniamotion.metaclass import MCMetaClass, DEFAULT_AXIS, DEFAULT_SERVO
from ingeniamotion.enums import OperationMode, SensorType, PhasingMode, GeneratorMode
from ingeniamotion.exceptions import IMTimeoutError


class Motion(metaclass=MCMetaClass):
    """Motion."""

    CONTROL_WORD_REGISTER = "DRV_STATE_CONTROL"
    OPERATION_MODE_REGISTER = "DRV_OP_CMD"
    OPERATION_MODE_DISPLAY_REGISTER = "DRV_OP_VALUE"
    POSITION_SET_POINT_REGISTER = "CL_POS_SET_POINT_VALUE"
    VELOCITY_SET_POINT_REGISTER = "CL_VEL_SET_POINT_VALUE"
    CURRENT_QUADRATURE_SET_POINT_REGISTER = "CL_CUR_Q_SET_POINT"
    CURRENT_DIRECT_SET_POINT_REGISTER = "CL_CUR_D_SET_POINT"
    VOLTAGE_QUADRATURE_SET_POINT_REGISTER = "CL_VOL_Q_SET_POINT"
    VOLTAGE_DIRECT_SET_POINT_REGISTER = "CL_VOL_D_SET_POINT"
    ACTUAL_POSITION_REGISTER = "CL_POS_FBK_VALUE"
    ACTUAL_VELOCITY_REGISTER = "CL_VEL_FBK_VALUE"
    ACTUAL_DIRECT_CURRENT_REGISTER = "CL_CUR_D_VALUE"
    ACTUAL_QUADRATURE_CURRENT_REGISTER = "CL_CUR_Q_VALUE"
    GENERATOR_FREQUENCY_REGISTER = "FBK_GEN_FREQ"
    GENERATOR_GAIN_REGISTER = "FBK_GEN_GAIN"
    GENERATOR_OFFSET_REGISTER = "FBK_GEN_OFFSET"
    GENERATOR_CYCLE_NUMBER_REGISTER = "FBK_GEN_CYCLES"
    GENERATOR_REARM_REGISTER = "FBK_GEN_REARM"

    STATUS_WORD_TARGET_REACHED_BIT = 0x800
    CONTROL_WORD_TARGET_LATCH_BIT = 0x200

    def __init__(self, motion_controller: "MotionController") -> None:
        self.mc = motion_controller
        self.logger = ingenialogger.get_logger(__name__)

    def target_latch(self, servo: str = DEFAULT_SERVO, axis: int = DEFAULT_AXIS) -> None:
        """Active target latch.

        Args:
            servo : servo alias to reference it. ``default`` by default.
            axis : servo axis. ``1`` by default.

        Raises:
            TypeError: If some read value has a wrong type.

        """
        control_word = self.mc.communication.get_register(
            self.CONTROL_WORD_REGISTER, servo=servo, axis=axis
        )
        if not isinstance(control_word, int):
            raise TypeError("Control word register value has to be a integer")
        new_control_word = control_word & (~self.CONTROL_WORD_TARGET_LATCH_BIT)
        self.mc.communication.set_register(
            self.CONTROL_WORD_REGISTER, new_control_word, servo=servo, axis=axis
        )
        new_control_word = control_word | self.CONTROL_WORD_TARGET_LATCH_BIT
        self.mc.communication.set_register(
            self.CONTROL_WORD_REGISTER, new_control_word, servo=servo, axis=axis
        )

    def set_operation_mode(
        self,
        operation_mode: Union[OperationMode, int],
        servo: str = DEFAULT_SERVO,
        axis: int = DEFAULT_AXIS,
    ) -> None:
        """Set operation mode to a target servo and axis.

        Args:
            operation_mode : operation mode, any of :class:`OperationMode`.
            servo : servo alias to reference it. ``default`` by default.
            axis : servo axis. ``1`` by default.

        """
        self.mc.communication.set_register(
            self.OPERATION_MODE_REGISTER, operation_mode, servo=servo, axis=axis
        )
        try:
            self.logger.debug(
                "Operation mode set to %s",
                OperationMode(operation_mode).name,
                axis=axis,
                drive=self.mc.servo_name(servo),
            )
        except ValueError:
            self.logger.debug(
                "Operation mode set to %s",
                operation_mode,
                axis=axis,
                drive=self.mc.servo_name(servo),
            )

    def get_operation_mode(
        self, servo: str = DEFAULT_SERVO, axis: int = DEFAULT_AXIS
    ) -> Union[OperationMode, int]:
        """Return current operation mode.

        Args:
            servo : servo alias to reference it. ``default`` by default.
            axis : servo axis. ``1`` by default.

        Returns:
            Return current operation mode.

        Raises:
            TypeError: If some read value has a wrong type.

        """
        operation_mode = self.mc.communication.get_register(
            self.OPERATION_MODE_DISPLAY_REGISTER, servo=servo, axis=axis
        )
        if not isinstance(operation_mode, (OperationMode, int)):
            raise TypeError("Operation mode value has to be an integer or OperationMode type.")
        try:
            return OperationMode(operation_mode)
        except ValueError:
            return operation_mode

    def motor_enable(self, servo: str = DEFAULT_SERVO, axis: int = DEFAULT_AXIS) -> None:
        """Enable motor.

        Args:
            servo : servo alias to reference it. ``default`` by default.
            axis : servo axis. ``1`` by default.

        Raises:
            ingenialink.exceptions.ILError: If the servo cannot enable the motor.

        """
        drive = self.mc.servos[servo]
        try:
            drive.enable(subnode=axis)
        except ILError as e:
            error_code, subnode, warning = self.mc.errors.get_last_buffer_error(
                servo=servo, axis=axis
            )
            error_id, _, _, error_msg = self.mc.errors.get_error_data(error_code, servo=servo)
            exception_type = type(e)
            raise exception_type("An error occurred enabling motor. Reason: {}".format(error_msg))

    def motor_disable(self, servo: str = DEFAULT_SERVO, axis: int = DEFAULT_AXIS) -> None:
        """Disable motor.

        Args:
            servo : servo alias to reference it. ``default`` by default.
            axis : servo axis. ``1`` by default.

        """
        try:
            is_motor_enabled = self.mc.configuration.is_motor_enabled(servo=servo, axis=axis)
        except ILError as e:
            self.logger.info(f"Unable to check if motor is enabled. Reason: {e}")
            return
        if is_motor_enabled:
            drive = self.mc.servos[servo]
            try:
                drive.disable(subnode=axis)
            except ILError as e:
                self.logger.info(f"Unable to disable the motor. Reason: {e}")

    def fault_reset(self, servo: str = DEFAULT_SERVO, axis: int = DEFAULT_AXIS) -> None:
        """Fault reset.

        Args:
            servo : servo alias to reference it. ``default`` by default.
            axis : servo axis. ``1`` by default.

        """
        drive = self.mc.servos[servo]
        try:
            drive.fault_reset(axis)
        except ILError as e:
            self.logger.info(f"Unable to perform a fault reset. Reason: {e}")

    def move_to_position(
        self,
        position: int,
        servo: str = DEFAULT_SERVO,
        axis: int = DEFAULT_AXIS,
        target_latch: bool = True,
        blocking: bool = False,
        error: int = 20,
        timeout: Optional[float] = None,
        interval: Optional[float] = None,
    ) -> None:
        """Set position set point to a target servo and axis, in counts.

        Args:
            position : target position, in counts.
            servo : servo alias to reference it. ``default`` by default.
            axis : servo axis. ``1`` by default.
            target_latch : if ``True`` does target latch at the end.
                ``True`` by default.
            blocking : if ``True``, the function is blocked until the
                target position is reached. ``False`` by default.
            error : If blocking is enabled, allowed error between actual
                position and target position, in counts.
            timeout : If blocking is enabled, how many seconds to wait
                for the servo to reach the target position, if ``None`` it
                will wait forever. ``None`` by default.
            interval : If blocking is enabled, interval of time between
                actual position reads, in seconds. ``None`` by default.

        Raises:
            TypeError: If position is not an int.
            IMTimeoutError: If the target position is not reached in time.

        """
        self.mc.communication.set_register(
            self.POSITION_SET_POINT_REGISTER, position, servo=servo, axis=axis
        )
        if target_latch:
            self.target_latch(servo, axis)
        if blocking:
            if not target_latch:
                self.logger.warning(
                    "Target latch is disabled. Target position may not be reached.",
                    axis=axis,
                    drive=self.mc.servo_name(servo),
                )
            self.wait_for_position(position, servo, axis, error, timeout, interval)

    def set_velocity(
        self,
        velocity: float,
        servo: str = DEFAULT_SERVO,
        axis: int = DEFAULT_AXIS,
        target_latch: bool = True,
        blocking: bool = False,
        error: float = 0.1,
        timeout: Optional[float] = None,
        interval: Optional[float] = None,
    ) -> None:
        """Set velocity set point to a target servo and axis, in rev/s.

        Args:
            velocity : target velocity, in rev/s.
            servo : servo alias to reference it. ``default`` by default.
            axis : servo axis. ``1`` by default.
            target_latch : if ``True`` does target latch at the end.
                ``True`` by default.
            blocking : if ``True``, the function is blocked until the
                target position is reached. ``False`` by default.
            error : If blocking is enabled, allowed error between
                actual velocity and target velocity, in rev/s.
            timeout : If blocking is enabled, how many seconds to wait
                for the servo to reach the target velocity, if ``None`` it
                will wait forever. ``None`` by default.
            interval : If blocking is enabled, interval of time between
                actual velocity reads, in seconds. ``None`` by default.
        Raises:
            TypeError: If velocity is not a float.
            IMTimeoutError: If the target velocity is not reached in time.

        """
        self.mc.communication.set_register(
            self.VELOCITY_SET_POINT_REGISTER, velocity, servo=servo, axis=axis
        )
        if target_latch:
            self.target_latch(servo, axis)
        if blocking:
            if not target_latch:
                self.logger.warning(
                    "Target latch is disabled. Target velocity may not be reached.",
                    axis=axis,
                    drive=self.mc.servo_name(servo),
                )
            self.wait_for_velocity(velocity, servo, axis, error, timeout, interval)

    def set_current_quadrature(
        self, current: float, servo: str = DEFAULT_SERVO, axis: int = DEFAULT_AXIS
    ) -> None:
        """Set quadrature current set point to a target servo and axis, in A.

        Args:
            current : target quadrature current, in A.
            servo : servo alias to reference it. ``default`` by default.
            axis : servo axis. ``1`` by default.

        Raises:
            TypeError: If current is not a float.

        """
        self.mc.communication.set_register(
            self.CURRENT_QUADRATURE_SET_POINT_REGISTER, current, servo=servo, axis=axis
        )

    def set_current_direct(
        self, current: float, servo: str = DEFAULT_SERVO, axis: int = DEFAULT_AXIS
    ) -> None:
        """Set direct current set point to a target servo and axis, in A.

        Args:
            current : target direct current, in A.
            servo : servo alias to reference it. ``default`` by default.
            axis : servo axis. ``1`` by default.

        Raises:
            TypeError: If current is not a float.

        """
        self.mc.communication.set_register(
            self.CURRENT_DIRECT_SET_POINT_REGISTER, current, servo=servo, axis=axis
        )

    def set_voltage_quadrature(
        self, voltage: float, servo: str = DEFAULT_SERVO, axis: int = DEFAULT_AXIS
    ) -> None:
        """Set quadrature voltage set point to a target servo and axis, in V.

        Args:
            voltage : target quadrature voltage, in V.
            servo : servo alias to reference it. ``default`` by default.
            axis : servo axis. ``1`` by default.

        Raises:
            TypeError: If voltage is not a float.

        """
        self.mc.communication.set_register(
            self.VOLTAGE_QUADRATURE_SET_POINT_REGISTER, voltage, servo=servo, axis=axis
        )

    def set_voltage_direct(
        self, voltage: float, servo: str = DEFAULT_SERVO, axis: int = DEFAULT_AXIS
    ) -> None:
        """Set direct voltage set point to a target servo and axis, in V.

        Args:
            voltage : target direct voltage, in V.
            servo : servo alias to reference it. ``default`` by default.
            axis : servo axis. ``1`` by default.

        Raises:
            TypeError: If voltage is not a float.

        """
        self.mc.communication.set_register(
            self.VOLTAGE_DIRECT_SET_POINT_REGISTER, voltage, servo=servo, axis=axis
        )

    def current_quadrature_ramp(
        self,
        target_value: float,
        time_s: float,
        servo: str = DEFAULT_SERVO,
        axis: int = DEFAULT_AXIS,
        init_value: float = 0,
        interval: Optional[float] = None,
    ) -> None:
        """Given a target value and a time in seconds, changes the current
        quadrature set-point linearly following a ramp. This function is
        blocked until target reached.

        Args:
            target_value : target value of the ramp.
            time_s : duration of the ramp, in seconds.
            servo : servo alias to reference it. ``default`` by default.
            axis : servo axis. ``1`` by default.
            init_value : initial value of the ramp. ``0`` by default.
            interval : time interval between register writes, in seconds.
                ``None`` by default, no interval.

        Raises:
            TypeError: If target_value or time_s is not a float.

        """
        for value in self.ramp_generator(init_value, target_value, time_s, interval):
            self.set_current_quadrature(value, servo=servo, axis=axis)

    def current_direct_ramp(
        self,
        target_value: float,
        time_s: float,
        servo: str = DEFAULT_SERVO,
        axis: int = DEFAULT_AXIS,
        init_value: float = 0,
        interval: Optional[float] = None,
    ) -> None:
        """
        Given a target value and a time in seconds, changes the current
        direct set-point linearly following a ramp. This function is
        blocked until target reached.

        Args:
            target_value : target value of the ramp.
            time_s : duration of the ramp, in seconds.
            servo : servo alias to reference it. ``default`` by default.
            axis : servo axis. ``1`` by default.
            init_value : initial value of the ramp. ``0`` by default.
            interval : time interval between register writes, in seconds.
                ``None`` by default, no interval.

        Raises:
            TypeError: If target_value or time_s is not a float.

        """
        for value in self.ramp_generator(init_value, target_value, time_s, interval):
            self.set_current_direct(value, servo=servo, axis=axis)

    def voltage_quadrature_ramp(
        self,
        target_value: float,
        time_s: float,
        servo: str = DEFAULT_SERVO,
        axis: int = DEFAULT_AXIS,
        init_value: float = 0,
        interval: Optional[float] = None,
    ) -> None:
        """
        Given a target value and a time in seconds, changes the voltage
        quadrature set-point linearly following a ramp. This function is
        blocked until target reached.

        Args:
            target_value : target value of the ramp.
            time_s : duration of the ramp, in seconds.
            servo : servo alias to reference it. ``default`` by default.
            axis : servo axis. ``1`` by default.
            init_value : initial value of the ramp. ``0`` by default.
            interval : time interval between register writes, in seconds.
                ``None`` by default, no interval.

        Raises:
            TypeError: If target_value or time_s is not a float.

        """
        for value in self.ramp_generator(init_value, target_value, time_s, interval):
            self.set_voltage_quadrature(value, servo=servo, axis=axis)

    def voltage_direct_ramp(
        self,
        target_value: float,
        time_s: float,
        servo: str = DEFAULT_SERVO,
        axis: int = DEFAULT_AXIS,
        init_value: float = 0,
        interval: Optional[float] = None,
    ) -> None:
        """
        Given a target value and a time in seconds, changes the voltage
        direct set-point linearly following a ramp. This function is
        blocked until target reached.

        Args:
            target_value : target value of the ramp.
            time_s : duration of the ramp, in seconds.
            servo : servo alias to reference it. ``default`` by default.
            axis : servo axis. ``1`` by default.
            init_value : initial value of the ramp. ``0`` by default.
            interval : time interval between register writes, in seconds.
                ``None`` by default, no interval.

        Raises:
            TypeError: If target_value or time_s is not a float.

        """
        for value in self.ramp_generator(init_value, target_value, time_s, interval):
            self.set_voltage_direct(value, servo=servo, axis=axis)

    @staticmethod
    def ramp_generator(
        init_v: float, final_v: float, total_t: float, interval: Optional[float] = None
    ) -> Generator[float, None, None]:
        slope = (final_v - init_v) / total_t
        init_time = time.time()
        yield init_v
        current_time = time.time()
        while current_time < init_time + total_t:
            yield slope * (current_time - init_time) + init_v
            if interval is not None:
                time.sleep(interval)
            current_time = time.time()
        yield final_v

    def get_actual_position(self, servo: str = DEFAULT_SERVO, axis: int = DEFAULT_AXIS) -> int:
        """
        Returns actual position register.

        Args:
            servo : servo alias to reference it. ``default`` by default.
            axis : servo axis. ``1`` by default.

        Returns:
            int: actual position value

        Raises:
            TypeError: If some read value has a wrong type.

        """
        actual_position = self.mc.communication.get_register(
            self.ACTUAL_POSITION_REGISTER, servo=servo, axis=axis
        )
        if not isinstance(actual_position, int):
            raise TypeError("Actual position value has to be an integer")
        return actual_position

    def get_actual_velocity(self, servo: str = DEFAULT_SERVO, axis: int = DEFAULT_AXIS) -> float:
        """
        Returns actual velocity register.

        Args:
            servo : servo alias to reference it. ``default`` by default.
            axis : servo axis. ``1`` by default.

        Returns:
            int: actual velocity value

        Raises:
            TypeError: If some read value has a wrong type.

        """
        actual_velocity = self.mc.communication.get_register(
            self.ACTUAL_VELOCITY_REGISTER, servo=servo, axis=axis
        )
        if not isinstance(actual_velocity, float):
            raise TypeError("Actual velocity value has to be an integer")
        return actual_velocity

    def get_actual_current_direct(
        self, servo: str = DEFAULT_SERVO, axis: int = DEFAULT_AXIS
    ) -> float:
        """
        Returns actual direct current register.

        Args:
            servo: servo alias to reference it. ``default`` by default.
            axis: servo axis. ``1`` by default.

        Returns:
            float: actual direct current value

        Raises:
            TypeError: If some read value has a wrong type.

        """
        actual_current_direct = self.mc.communication.get_register(
            self.ACTUAL_DIRECT_CURRENT_REGISTER, servo=servo, axis=axis
        )
        if not isinstance(actual_current_direct, float):
            raise TypeError("Actual current direct value has to be a float")
        return actual_current_direct

    def get_actual_current_quadrature(
        self, servo: str = DEFAULT_SERVO, axis: int = DEFAULT_AXIS
    ) -> float:
        """
        Returns actual quadrature current register.

        Args:
            servo (str): servo alias to reference it. ``default`` by default.
            axis (int): servo axis. ``1`` by default.

        Returns:
            float: actual quadrature current value

        Raises:
            TypeError: If some read value has a wrong type.

        """
        actual_current_quadrature = self.mc.communication.get_register(
            self.ACTUAL_QUADRATURE_CURRENT_REGISTER, servo=servo, axis=axis
        )
        if not isinstance(actual_current_quadrature, float):
            raise TypeError("Actual current quadrature value has to be a float")
        return actual_current_quadrature

    def wait_for_position(
        self,
        position: int,
        servo: str = DEFAULT_SERVO,
        axis: int = DEFAULT_AXIS,
        error: int = 20,
        timeout: Optional[float] = None,
        interval: Optional[float] = None,
    ) -> None:
        """
        Wait until actual position is equal to a target position, with an error.

        Args:
            position : target position, in counts.
            servo : servo alias to reference it. ``default`` by default.
            axis : servo axis. ``1`` by default.
            error : allowed error between actual position and target
                position, in counts.
            timeout : how many seconds to wait for the servo to reach the
                target position, if ``None`` it will wait forever .
                ``None`` by default.
            interval : interval of time between actual position reads,
                in seconds. ``None`` by default.

        Raises:
            IMTimeoutError: If the target position is not reached in time.

        """
        target_reached = False
        init_time = time.time()
        self.logger.debug(
            "Wait for position %s", position, axis=axis, drive=self.mc.servo_name(servo)
        )
        while not target_reached:
            if interval:
                time.sleep(interval)
            curr_position = self.get_actual_position(servo=servo, axis=axis)
            target_reached = abs(position - curr_position) < abs(error)
            if timeout and (init_time + timeout) < time.time():
                target_reached = True
                self.logger.warning(
                    "Timeout: position %s was not reached",
                    position,
                    axis=axis,
                    drive=self.mc.servo_name(servo),
                )
                raise IMTimeoutError("Position was not reached in time")

    def wait_for_velocity(
        self,
        velocity: float,
        servo: str = DEFAULT_SERVO,
        axis: int = DEFAULT_AXIS,
        error: float = 0.1,
        timeout: Optional[float] = None,
        interval: Optional[float] = None,
    ) -> None:
        """
        Wait until actual velocity is equal to a target velocity, with an error.

        Args:
            velocity : target velocity, in rev/s.
            servo : servo alias to reference it. ``default`` by default.
            axis : servo axis. ``1`` by default.
            error : allowed error between actual velocity and target
                velocity, in rev/s.
            timeout : how many seconds to wait for the servo to reach the
                target velocity, if ``None`` it will wait forever.
                ``None`` by default.
            interval : interval of time between actual velocity reads,
                in seconds. ``None`` by default.

        Raises:
            IMTimeoutError: If the target velocity is not reached in time.

        """
        target_reached = False
        init_time = time.time()
        self.logger.debug(
            "Wait for velocity %s", velocity, axis=axis, drive=self.mc.servo_name(servo)
        )
        while not target_reached:
            if interval:
                time.sleep(interval)
            curr_velocity = self.get_actual_velocity(servo=servo, axis=axis)
            target_reached = abs(velocity - curr_velocity) < abs(error)
            if timeout and (init_time + timeout) < time.time():
                target_reached = True
                self.logger.warning(
                    "Timeout: velocity %s was not reached",
                    velocity,
                    axis=axis,
                    drive=self.mc.servo_name(servo),
                )
                raise IMTimeoutError("Velocity was not reached in time")

    def set_internal_generator_configuration(
        self, op_mode: OperationMode, servo: str = DEFAULT_SERVO, axis: int = DEFAULT_AXIS
    ) -> None:
        """
        Set internal generator configuration.

        .. note::
            This functions affects the following drive registers: **motor pair poles**,
            **operation mode**, **phasing mode** and **commutation feedback**.
            This is an advanced-user oriented method, you could lose your drive
            configuration, use it at your own risk.

        Args:
            op_mode : select Current or Voltage operation mode
             for internal generator configuration.
            servo : servo alias to reference it. ``default`` by default.
            axis : servo axis. ``1`` by default.

        Raises:
            ValueError: If operation mode is not set to Current or Voltage.

        """
        if op_mode not in [OperationMode.CURRENT, OperationMode.VOLTAGE]:
            raise ValueError("Operation mode must be Current or Voltage")
        self.set_operation_mode(op_mode, servo=servo, axis=axis)
        self.mc.configuration.set_motor_pair_poles(1, servo=servo, axis=axis)
        self.mc.configuration.set_phasing_mode(PhasingMode.NO_PHASING, servo=servo, axis=axis)
        if op_mode == OperationMode.CURRENT:
            self.set_current_quadrature(0, servo=servo, axis=axis)
            self.set_current_direct(0, servo=servo, axis=axis)
        elif op_mode == OperationMode.VOLTAGE:
            self.set_voltage_quadrature(0, servo=servo, axis=axis)
            self.set_voltage_direct(0, servo=servo, axis=axis)
        self.mc.configuration.set_commutation_feedback(SensorType.INTGEN, servo=servo, axis=axis)

    def internal_generator_saw_tooth_move(
        self,
        direction: int,
        cycles: int,
        frequency: float,
        gain: float = 1.0,
        servo: str = DEFAULT_SERVO,
        axis: int = DEFAULT_AXIS,
    ) -> None:
        """
        Move motor in internal generator configuration with generator mode saw tooth.

        Args:
            direction : ``1`` for positive direction and
             ``-1`` for negative direction.
            cycles : movement cycles.
            frequency : cycles for second.
            gain: positive generator gain.
            servo : servo alias to reference it. ``default`` by default.
            axis : servo axis. ``1`` by default.

        Raises:
            TypeError: If direction, cycles or frequency is not of the correct type.
            ValueError: If gain is not positive.

        """
        if gain < 0:
            raise ValueError("Gain should be positive")
        self.mc.configuration.set_generator_mode(GeneratorMode.SAW_TOOTH, servo=servo, axis=axis)
        self.mc.communication.set_register(
            self.GENERATOR_FREQUENCY_REGISTER, frequency, servo=servo, axis=axis
        )
        if direction > 0:
            self.mc.communication.set_register(
                self.GENERATOR_GAIN_REGISTER, gain, servo=servo, axis=axis
            )
            self.mc.communication.set_register(
                self.GENERATOR_OFFSET_REGISTER, 0, servo=servo, axis=axis
            )
        else:
            self.mc.communication.set_register(
                self.GENERATOR_GAIN_REGISTER, -gain, servo=servo, axis=axis
            )
            self.mc.communication.set_register(
                self.GENERATOR_OFFSET_REGISTER, gain, servo=servo, axis=axis
            )

        self.mc.communication.set_register(
            self.GENERATOR_CYCLE_NUMBER_REGISTER, cycles, servo=servo, axis=axis
        )
        self.mc.communication.set_register(self.GENERATOR_REARM_REGISTER, 1, servo=servo, axis=axis)

    def internal_generator_constant_move(
        self, offset: int, servo: str = DEFAULT_SERVO, axis: int = DEFAULT_AXIS
    ) -> None:
        """
        Move motor in internal generator configuration with generator mode constant.

        Args:
            offset : internal generator offset.
            servo : servo alias to reference it. ``default`` by default.
            axis : servo axis. ``1`` by default.

        Raises:
            TypeError: If offset is not an int.

        """
        self.mc.configuration.set_generator_mode(GeneratorMode.CONSTANT, servo=servo, axis=axis)
        self.mc.communication.set_register(self.GENERATOR_GAIN_REGISTER, 0, servo=servo, axis=axis)
        self.mc.communication.set_register(
            self.GENERATOR_OFFSET_REGISTER, offset, servo=servo, axis=axis
        )



import os
from typing import TYPE_CHECKING, Dict, Optional, Tuple, Union

from ingenialink import CAN_BAUDRATE
from ingenialink.eoe.network import EoENetwork
from ingenialink.ethernet.network import EthernetNetwork
from ingenialink.canopen.network import CanopenNetwork
from ingenialink.ethercat.network import EthercatNetwork
from ingenialink.register import Register
from ingenialink.enums.register import REG_ACCESS, REG_DTYPE
from ingenialink.dictionary import SubnodeType
import ingenialogger

from ingeniamotion.exceptions import IMRegisterNotExist, IMException
from ingeniamotion.metaclass import MCMetaClass, DEFAULT_AXIS, DEFAULT_SERVO
from ingeniamotion.enums import COMMUNICATION_TYPE


if TYPE_CHECKING:
    from ingeniamotion.motion_controller import MotionController

logger = ingenialogger.get_logger(__name__)


class Information(metaclass=MCMetaClass):
    """Information."""

    def __init__(self, motion_controller: "MotionController"):
        self.mc = motion_controller

    def register_info(
        self,
        register: str,
        axis: int = DEFAULT_AXIS,
        servo: str = DEFAULT_SERVO,
    ) -> Register:
        """Return register object.

        Args:
            register : register UID.
            axis : servo axis. ``1`` by default.
            servo : servo alias to reference it. ``default`` by default.

        Returns:
            Register object.

        Raises:
            IMRegisterNotExist: If register does not exist in dictionary.

        """
        drive = self.mc.servos[servo]
        try:
            return drive.dictionary.registers(axis)[register]
        except KeyError:
            raise IMRegisterNotExist(
                "Register: {} axis: {} not exist in dictionary".format(register, axis)
            )

    def register_type(
        self,
        register: str,
        axis: int = DEFAULT_AXIS,
        servo: str = DEFAULT_SERVO,
    ) -> REG_DTYPE:
        """Return register dtype.

        Args:
            register : register UID.
            axis : servo axis. ``1`` by default.
            servo : servo alias to reference it. ``default`` by default.

        Returns:
            Register dtype.

        Raises:
            IMRegisterNotExist: If register does not exist in dictionary.

        """
        register_obj = self.register_info(register, axis=axis, servo=servo)
        return register_obj.dtype

    def register_access(
        self,
        register: str,
        axis: int = DEFAULT_AXIS,
        servo: str = DEFAULT_SERVO,
    ) -> REG_ACCESS:
        """Return register access.

        Args:
            register : register UID.
            axis : servo axis. ``1`` by default.
            servo : servo alias to reference it. ``default`` by default.

        Returns:
            Register access.

        Raises:
            IMRegisterNotExist: If register does not exist in dictionary.

        """
        register_obj = self.register_info(register, axis=axis, servo=servo)
        return register_obj.access

    def register_range(
        self,
        register: str,
        axis: int = DEFAULT_AXIS,
        servo: str = DEFAULT_SERVO,
    ) -> Union[Tuple[None, None], Tuple[int, int], Tuple[float, float], Tuple[str, str]]:
        """Return register range.

        Args:
            register : register UID.
            axis : servo axis. ``1`` by default.
            servo : servo alias to reference it. ``default`` by default.

        Returns:
            Register range, minimum and maximum.

        Raises:
            IMRegisterNotExist: If register does not exist in dictionary.

        """
        register_obj = self.register_info(register, axis=axis, servo=servo)
        return register_obj.range

    def register_exists(
        self,
        register: str,
        axis: int = DEFAULT_AXIS,
        servo: str = DEFAULT_SERVO,
    ) -> bool:
        """Check if register exists in dictionary.

        Args:
            register : register UID.
            axis : servo axis. ``1`` by default.
            servo : servo alias to reference it. ``default`` by default.

        Returns:
            ``True`` if register exists, else ``False``.

        """
        drive = self.mc.servos[servo]
        return register in drive.dictionary.registers(axis)

    def get_product_name(self, alias: str = DEFAULT_SERVO) -> Optional[str]:
        """Get the product name of the drive.

        Args:
            alias: alias of the servo.

        Returns:
            If it exists for example: "EVE-NET-E", "CAP-NET-E", etc.
        """
        drive = self.mc.servos[alias]
        product_name = drive.dictionary.part_number
        if not product_name is None:
            return f"{product_name}"
        return None

    def get_node_id(self, alias: str = DEFAULT_SERVO) -> int:
        """Get the node ID for CANopen communications.

        Args:
            alias: alias of the servo.

        Returns:
            Node ID of the drive.
        """
        net = self.mc._get_network(alias)
        drive = self.mc.servos[alias]
        if isinstance(net, CanopenNetwork):
            return int(drive.target)
        else:
            raise IMException("You need a CANopen communication to use this function")

    def get_baudrate(self, alias: str = DEFAULT_SERVO) -> CAN_BAUDRATE:
        """Get the baudrate of target servo

        Args:
            alias: alias of the servo.

        Returns:
            Baudrate of the drive.
        """
        net = self.mc._get_network(alias)
        if isinstance(net, CanopenNetwork):
            return CAN_BAUDRATE(net.baudrate)
        raise IMException(f"The servo {alias} is not a CANopen device.")

    def get_ip(self, alias: str = DEFAULT_SERVO) -> str:
        """Get the IP for Ethernet communications.

        Args:
            alias: alias of the servo.

        Returns:
            IP of the drive.
        """
        net = self.mc._get_network(alias)
        drive = self.mc.servos[alias]
        if isinstance(net, EthernetNetwork):
            return str(drive.target)
        else:
            raise IMException("You need an Ethernet communication to use this function")

    def get_slave_id(self, alias: str = DEFAULT_SERVO) -> int:
        """Get the slave ID for EoE communications.

        Args:
            alias: alias of the servo.

        Returns:
            Slave ID of the drive.
        """
        net = self.mc._get_network(alias)
        drive = self.mc.servos[alias]
        if isinstance(net, EoENetwork) and isinstance(drive.target, str):
            return int(net._configured_slaves[drive.target])
        else:
            raise IMException("You need a CANopen communication to use this function")

    def get_name(self, alias: str = DEFAULT_SERVO) -> str:
        """Get the drive's name.

        Args:
            alias: Alias of the servo.
        Returns:
            The name of the drive.
        """
        drive = self.mc.servos[alias]
        drive_name = drive.name
        return f"{drive_name}"

    def get_communication_type(self, alias: str = DEFAULT_SERVO) -> COMMUNICATION_TYPE:
        """Get the connected drive's communication type.

        Args:
            alias: alias of the connected drive.

        Returns:
            CANopen, Ethernet, or EtherCAT.

        """
        drive_network = self.mc._get_network(alias)
        if isinstance(drive_network, CanopenNetwork):
            communication_type = COMMUNICATION_TYPE.Canopen
        elif isinstance(drive_network, EthernetNetwork):
            communication_type = COMMUNICATION_TYPE.Ethernet
        elif isinstance(drive_network, (EoENetwork, EthercatNetwork)):
            communication_type = COMMUNICATION_TYPE.Ethercat
        return communication_type

    def get_full_name(self, alias: str = DEFAULT_SERVO) -> str:
        """Return the full name of the drive [Product name] [Name] ([Target]).

        Args:
            alias: Drive alias.

        Returns:
            Full name.
        """
        prod_name = self.get_product_name(alias)
        name = self.get_name(alias)
        full_name = f"{prod_name} - {name}"
        net = self.mc._get_network(alias)
        if isinstance(net, EthernetNetwork):
            ip = self.get_ip(alias)
            full_name = f"{full_name} ({ip})"
        return full_name

    def get_subnodes(self, alias: str = DEFAULT_SERVO) -> Dict[int, SubnodeType]:
        """Return a dictionary with the subnodes IDs as keys and their type as values.

        Args:
            alias: Drive alias.

        Returns:
            Dictionary of subnode ids and their type.
        """
        drive = self.mc.servos[alias]
        return drive.subnodes

    def get_categories(self, alias: str) -> Dict[str, str]:
        """Return dictionary categories instance.

        Args:
            alias: Drive alias.

        Returns:
            Categories instance.
        """
        drive = self.mc.servos[alias]
        dictionary_categories = drive.dictionary.categories
        if not dictionary_categories:
            raise IMException("Dictionary categories are not defined.")
        category_ids = dictionary_categories.category_ids
        categories: Dict[str, str] = {}
        for cat_id in category_ids:
            categories[cat_id] = dictionary_categories.labels(cat_id)["en_US"]
        return categories

    def get_dictionary_file_name(self, alias: str) -> str:
        """Return dictionary file name.

        Args:
            alias: Drive alias.

        Returns:
            Dictionary file name.
        """
        drive = self.mc.servos[alias]
        return str(os.path.basename(drive.dictionary.path))

    def get_encoded_image_from_dictionary(self, alias: str) -> Optional[str]:
        """Get the encoded product image from a drive dictionary.
        This function reads a dictionary of a drive, and it parses whether the dictionary file has a
        DriveImage tag and its content.
        Args:
            alias: Alias of the drive.
        Returns:
            The encoded image or NoneType object.
        """
        drive = self.mc.servos[alias]
        return drive.dictionary.image



import math
from enum import IntEnum
from typing import TYPE_CHECKING, Optional

import ingenialogger

from ingeniamotion.exceptions import IMTimeoutError
from ingeniamotion.wizard_tests.base_test import BaseTest, TestError
from ingeniamotion.enums import SensorType, OperationMode, SeverityLevel

if TYPE_CHECKING:
    from ingeniamotion.motion_controller import MotionController


class DCFeedbacksResolutionTest(BaseTest):
    MOVEMENT_ERROR_FACTOR = 0.05
    DEFAULT_PROFILE_MAX_VEL = 0.3
    DEFAULT_VELOCITY_PID = {"kp": 0.1, "ki": 10, "kd": 0}
    POSITION_TUNE_BW = 1
    MOVEMENT_EXTRA_TIME_FACTOR = 1.5
    MOVEMENT_TIMEOUT = MOVEMENT_EXTRA_TIME_FACTOR / DEFAULT_PROFILE_MAX_VEL
    OPERATION_MODE = OperationMode.PROFILE_POSITION

    PID_LOG_MSG = "Kp = {kp}, Ki = {ki} and Kd = {kd}"

    class ResultType(IntEnum):
        SUCCESS = 0

    result_description = {
        ResultType.SUCCESS: "Feedback resolution test has been executed",
    }

    BACKUP_REGISTERS = [
        "CL_POS_FBK_SENSOR",
        "CL_VEL_FBK_SENSOR",
        "CL_AUX_FBK_SENSOR",
        "DRV_OP_CMD",
        "CL_VEL_PID_KP",
        "CL_VEL_PID_KI",
        "CL_VEL_PID_KD",
        "CL_POS_PID_KP",
        "CL_POS_PID_KI",
        "CL_POS_PID_KD",
        "PROF_MAX_VEL",
    ]

    feedback_resolution: int

    def __init__(
        self,
        mc: "MotionController",
        sensor: SensorType,
        servo: str,
        axis: int,
        kp: Optional[float] = None,
        ki: Optional[float] = None,
        kd: Optional[float] = None,
    ):
        if sensor == SensorType.HALLS:
            raise NotImplementedError("This test is not implemented for Hall sensor")
        super().__init__()
        self.mc = mc
        self.sensor = sensor
        self.servo = servo
        self.axis = axis
        self.backup_registers_names = self.BACKUP_REGISTERS
        self.test_velocity_pid = self.DEFAULT_VELOCITY_PID.copy()
        if kp is not None:
            self.test_velocity_pid["kp"] = kp
            self.test_velocity_pid["ki"] = ki or 0
            self.test_velocity_pid["kd"] = kd or 0
        self.logger = ingenialogger.get_logger(__name__, axis=axis, drive=mc.servo_name(servo))

    @BaseTest.stoppable
    def setup(self) -> None:
        self.mc.motion.motor_disable(servo=self.servo, axis=self.axis)
        self.logger.info("Motor disable")
        self.feedback_resolution = self.mc.configuration.get_feedback_resolution(
            self.sensor, servo=self.servo, axis=self.axis
        )
        self.mc.configuration.set_velocity_feedback(self.sensor, servo=self.servo, axis=self.axis)
        self.mc.configuration.set_position_feedback(self.sensor, servo=self.servo, axis=self.axis)
        if self.sensor == SensorType.BISSC2:
            self.mc.configuration.set_auxiliar_feedback(
                SensorType.ABS1, servo=self.servo, axis=self.axis
            )
            self.logger.info(
                f"Set velocity and position feedbacks to {self.sensor.name}"
                f" and axuiliar to {SensorType.ABS1.name}"
            )
        else:
            self.mc.configuration.set_auxiliar_feedback(
                self.sensor, servo=self.servo, axis=self.axis
            )
            self.logger.info(f"Set velocity, position and auxiliar feedbacks to {self.sensor.name}")
        self.mc.configuration.set_velocity_pid(
            **self.test_velocity_pid, servo=self.servo, axis=self.axis
        )
        self.logger.info(f"Velocity PID set to {self.PID_LOG_MSG.format(**self.test_velocity_pid)}")
        position_kp = 2.0 * math.pi * self.POSITION_TUNE_BW / self.feedback_resolution
        self.mc.configuration.set_position_pid(kp=position_kp, servo=self.servo, axis=self.axis)
        self.logger.info(
            f"Position PID set to {self.PID_LOG_MSG.format(kp=position_kp, ki=0, kd=0)}"
        )
        self.mc.configuration.set_max_profile_velocity(
            self.DEFAULT_PROFILE_MAX_VEL, servo=self.servo, axis=self.axis
        )
        self.logger.info(f"Maximum profile velocity set to {self.DEFAULT_PROFILE_MAX_VEL}")
        self.mc.motion.set_operation_mode(self.OPERATION_MODE, servo=self.servo, axis=self.axis)
        self.logger.info(f"Set operation mode to {self.OPERATION_MODE.name}")

    @BaseTest.stoppable
    def loop(self) -> ResultType:
        initial_pos = self.mc.motion.get_actual_position(servo=self.servo, axis=self.axis)
        self.mc.motion.move_to_position(initial_pos, servo=self.servo, axis=self.axis)
        self.mc.motion.motor_enable(servo=self.servo, axis=self.axis)
        self.logger.info("Motor enable")
        try:
            self.logger.info(f"Try to reach position {initial_pos + self.feedback_resolution}")
            self.mc.motion.move_to_position(
                initial_pos + self.feedback_resolution,
                servo=self.servo,
                axis=self.axis,
                blocking=True,
                error=int(self.feedback_resolution * self.MOVEMENT_ERROR_FACTOR),
                timeout=self.MOVEMENT_TIMEOUT,
            )
        except IMTimeoutError as e:
            final_position = self.mc.motion.get_actual_position(servo=self.servo, axis=self.axis)
            if (
                abs(final_position - initial_pos)
                > self.feedback_resolution * self.MOVEMENT_ERROR_FACTOR
            ):
                raise TestError(e) from e
            raise TestError(
                "ERROR: No movement detected. Please, review your feedback configuration & wiring"
            ) from e
        self.mc.motion.motor_disable(servo=self.servo, axis=self.axis)
        return self.ResultType.SUCCESS

    def teardown(self) -> None:
        self.mc.motion.motor_disable(servo=self.servo, axis=self.axis)
        self.logger.info("Motor disable")

    def get_result_msg(self, output: ResultType) -> str:
        description = self.result_description.get(output)
        if description is None:
            raise NotImplementedError("Result description not implemented")
        return description

    def get_result_severity(self, output: ResultType) -> SeverityLevel:
        if output == self.ResultType.SUCCESS:
            return SeverityLevel.SUCCESS
        return SeverityLevel.FAIL



import math
import time
from enum import IntEnum
from typing import TYPE_CHECKING, Optional

import ingenialogger
from ingenialink.exceptions import ILTimeoutError, ILStateError, ILIOError

if TYPE_CHECKING:
    from ingeniamotion import MotionController

from ingeniamotion.enums import OperationMode, SensorType, SeverityLevel
from ingeniamotion.exceptions import IMRegisterNotExist
from ingeniamotion.wizard_tests.base_test import BaseTest, TestError
from ingeniamotion.exceptions import IMRegisterNotExist
from ingeniamotion.enums import OperationMode, SeverityLevel, SensorType, CommutationMode


class Feedbacks(BaseTest):
    """Feedbacks Wizard Class description."""

    class ResultType(IntEnum):
        SUCCESS = 0
        RESOLUTION_ERROR = -1
        SYMMETRY_ERROR = -2
        POS_VEL_RATIO_ERROR = -4

    class Polarity(IntEnum):
        NORMAL = 0
        REVERSED = 1

    result_description = {
        ResultType.SUCCESS: "Feedback test pass successfully",
        ResultType.RESOLUTION_ERROR: "Feedback has a resolution error",
        ResultType.SYMMETRY_ERROR: "Feedback has a symmetry error",
        ResultType.POS_VEL_RATIO_ERROR: "Position to velocity sensor ratio cannot be different "
        "than 1 when both feedback sensors are the same.",
    }

    # Aux constants
    FEEDBACK_TOLERANCE = 17
    TEST_FREQUENCY = 0.4
    TIME_BETWEEN_MOVEMENT = 0.5
    PERCENTAGE_CURRENT_USED = 0.8
    LOW_PASS_FILTER = 1

    FAIL_MSG_MISMATCH = "A mismatch in resolution has been detected."

    VELOCITY_FEEDBACK_FILTER_1_TYPE_REGISTER = "CL_VEL_FBK_FILTER1_TYPE"
    VELOCITY_FEEDBACK_FILTER_1_FREQUENCY_REGISTER = "CL_VEL_FBK_FILTER1_FREQ"
    RATED_CURRENT_REGISTER = "MOT_RATED_CURRENT"
    MAXIMUM_CONTINUOUS_CURRENT_DRIVE_PROTECTION = "DRV_PROT_MAN_MAX_CONT_CURRENT_VALUE"
    POSITIONING_OPTION_CODE_REGISTER = "PROF_POS_OPTION_CODE"
    MAX_POSITION_RANGE_LIMIT_REGISTER = "CL_POS_REF_MAX_RANGE"
    MIN_POSITION_RANGE_LIMIT_REGISTER = "CL_POS_REF_MIN_RANGE"

    BACKUP_REGISTERS = [
        "CL_POS_FBK_SENSOR",
        "MOT_PAIR_POLES",
        "DRV_OP_CMD",
        "CL_CUR_Q_SET_POINT",
        "CL_CUR_D_SET_POINT",
        "FBK_GEN_MODE",
        "FBK_GEN_FREQ",
        "FBK_GEN_GAIN",
        "FBK_GEN_OFFSET",
        "COMMU_ANGLE_SENSOR",
        "FBK_GEN_CYCLES",
        "FBK_SSI2_POS_POLARITY",
        "COMMU_PHASING_MODE",
        "MOT_COMMU_MOD",
        "CL_AUX_FBK_SENSOR",
        "ERROR_DIGENC_AGAINST_HALL_OPTION",
        "CL_VEL_FOLLOWING_OPTION",
        "ERROR_VEL_OUT_LIMITS_OPTION",
        "ERROR_POS_OUT_LIMITS_OPTION",
        "ERROR_POS_FOLLOWING_OPTION",
        "CL_VEL_FBK_SENSOR",
        "COMMU_ANGLE_REF_SENSOR",
        "CL_VEL_FBK_FILTER1_TYPE",
        "CL_VEL_FBK_FILTER1_FREQ",
    ]

    OPTIONAL_BACKUP_REGISTERS = [
        "COMMU_ANGLE_INTEGRITY1_OPTION",
        "COMMU_ANGLE_INTEGRITY2_OPTION",
        POSITIONING_OPTION_CODE_REGISTER,
        MAX_POSITION_RANGE_LIMIT_REGISTER,
        MIN_POSITION_RANGE_LIMIT_REGISTER,
    ]

    FEEDBACK_POLARITY_REGISTER: str

    SENSOR_TYPE_FEEDBACK_TEST: SensorType

    def __init__(self, mc: "MotionController", servo: str, axis: int) -> None:
        super().__init__()
        self.mc = mc
        self.servo = servo
        self.axis = axis
        self.sensor = self.SENSOR_TYPE_FEEDBACK_TEST
        self.logger = ingenialogger.get_logger(__name__, axis=axis, drive=mc.servo_name(servo))
        self.feedback_resolution: Optional[int] = None
        self.pair_poles: Optional[int] = None
        self.pos_vel_same_feedback = False
        self.resolution_multiplier = 1.0
        self.test_frequency = self.TEST_FREQUENCY
        self.backup_registers_names = self.BACKUP_REGISTERS.copy()
        self.optional_backup_registers_names = self.OPTIONAL_BACKUP_REGISTERS.copy()
        self.suggested_registers = {}

    @BaseTest.stoppable
    def check_feedback_tolerance(
        self, error: float, error_msg: str, error_type: ResultType
    ) -> ResultType:
        if error > self.FEEDBACK_TOLERANCE:
            error_advice = "Please, review your feedback & motor pair poles settings"
            self.logger.error("%s %s", error_msg, error_advice)
            return error_type
        return self.ResultType.SUCCESS

    @BaseTest.stoppable
    def check_symmetry(self, positive: float, negative: float) -> ResultType:
        self.logger.info("SYMMETRY CHECK")
        if not isinstance(self.feedback_resolution, int):
            raise TypeError("Feedbacks has to be set before symetry checking.")
        error = 100 * abs(positive + negative) / self.feedback_resolution
        self.logger.info("Detected symmetry mismatch of: %.3f%%", error)
        error_msg = (
            "ERROR: A mismatch in resolution has been "
            "detected between positive and negative direction."
        )
        return self.check_feedback_tolerance(error, error_msg, self.ResultType.SYMMETRY_ERROR)

    @BaseTest.stoppable
    def check_polarity(self, displacement: float) -> Polarity:
        self.logger.info("POLARITY CHECK")
        polarity = self.Polarity.NORMAL if displacement > 0 else self.Polarity.REVERSED
        self.logger.info("Feedback polarity detected: %s", polarity.name)
        return polarity

    @BaseTest.stoppable
    def check_resolution(self, displacement: float) -> ResultType:
        if self.pair_poles is None:
            raise TypeError("Pair poles has to be set before resolution checking.")
        if self.feedback_resolution is None:
            raise TypeError("Feedback resolution has to be set before resolution checking.")
        displacement = displacement * self.pair_poles
        self.logger.info("RESOLUTION CHECK")
        self.logger.info("Theoretical resolution: %.0f", self.feedback_resolution)
        self.logger.info("Detected resolution (pos): %.0f", abs(displacement))
        displacement_value = abs(self.feedback_resolution - abs(displacement))
        error = 100 * displacement_value / self.feedback_resolution
        self.logger.info("Detected mismatch of: %.3f%%", error)
        error_msg = (
            "ERROR: The detected feedback resolution does not "
            "match with the specified in the configuration."
        )
        return self.check_feedback_tolerance(error, error_msg, self.ResultType.RESOLUTION_ERROR)

    @BaseTest.stoppable
    def feedback_setting(self) -> None:
        # First set all feedback to feedback in test, so there won't be
        # more than 5 feedback at the same time
        self.mc.configuration.set_commutation_feedback(
            self.sensor, servo=self.servo, axis=self.axis
        )
        self.mc.configuration.set_reference_feedback(self.sensor, servo=self.servo, axis=self.axis)
        self.mc.configuration.set_velocity_feedback(self.sensor, servo=self.servo, axis=self.axis)
        self.mc.configuration.set_position_feedback(self.sensor, servo=self.servo, axis=self.axis)
        self.mc.configuration.set_auxiliar_feedback(self.sensor, servo=self.servo, axis=self.axis)
        # Set Polarity to 0
        self.mc.communication.set_register(
            self.FEEDBACK_POLARITY_REGISTER, self.Polarity.NORMAL, servo=self.servo, axis=self.axis
        )
        # Depending on the type of the feedback, calculate the correct
        # feedback resolution
        self.feedback_resolution = self.mc.configuration.get_feedback_resolution(
            self.sensor, servo=self.servo, axis=self.axis
        )

    @BaseTest.stoppable
    def reaction_codes_to_warning(self) -> None:
        # TODO Add function in errors to disable errors
        # set velocity and position following errors to WARNING = 1
        # ignore failed writes
        following_error_uids = [
            "ERROR_DIGENC_AGAINST_HALL_OPTION",
            "ERROR_DIGHALL_SEQ_OPTION",
            "CL_VEL_FOLLOWING_OPTION",
            "ERROR_VEL_OUT_LIMITS_OPTION",
            "ERROR_POS_OUT_LIMITS_OPTION",
            "ERROR_POS_FOLLOWING_OPTION",
            "COMMU_ANGLE_INTEGRITY1_OPTION",
            "COMMU_ANGLE_INTEGRITY2_OPTION",
        ]
        for following_error_uid in following_error_uids:
            try:
                self.mc.communication.set_register(
                    following_error_uid, 1, servo=self.servo, axis=self.axis
                )
            except IMRegisterNotExist as e:
                self.logger.warning(e)

    @BaseTest.stoppable
    def suggest_polarity(self, pol: Polarity) -> None:
        if not isinstance(self.FEEDBACK_POLARITY_REGISTER, str):
            raise TypeError("Feedback polarity register has to be set before polarity suggestion.")
        polarity_uid = self.FEEDBACK_POLARITY_REGISTER
        self.suggested_registers[polarity_uid] = pol

    @BaseTest.stoppable
    def setup(self) -> None:
        # Prerequisites:
        #  - Motor & Feedbacks configured (Pair poles & rated current are used)
        #  - Current control loop tuned
        #  - Feedback reaction codes to WARNING
        # Protection to avoid any unwanted movement
        self.mc.motion.motor_disable(servo=self.servo, axis=self.axis)
        self.logger.info("CONFIGURATION OF THE TEST")
        # Set commutation modulation to sinusoidal
        self.mc.configuration.set_commutation_mode(
            CommutationMode.SINUSOIDAL, servo=self.servo, axis=self.axis
        )
        # Set positioning mode to NO LIMITS
        self.__set_positioning_register_values()
        # Default resolution multiplier
        self.__set_resolution_multiplier()
        # Read pole pairs and set to 1 for an electrical revolution
        self.pair_poles = self.mc.configuration.get_motor_pair_poles(
            servo=self.servo, axis=self.axis
        )
        # For each feedback on motor side we should repeat this test using the
        # feedback as position sensor. The polarity of the feedback must be set
        # also to normal at the beginning. All feedback are set to the same,
        # in order to avoid feedback configuration
        # error (wizard_tests series can only support 4 feedback at the
        # same time)
        self.feedback_setting()

        self.mc.motion.set_internal_generator_configuration(
            OperationMode.CURRENT, servo=self.servo, axis=self.axis
        )
        self.logger.info("Set pair poles to 1")
        self.logger.info("Mode of operation set to Current mode")
        self.logger.info("Set phasing mode to No phasing")
        self.logger.info("Target quadrature current set to zero")
        self.logger.info("Target direct current set to zero")
        self.logger.info("Commutation feedback set to Internal Generator")

        # set velocity and position following errors to WARNING = 1
        self.reaction_codes_to_warning()

    def __set_positioning_register_values(self) -> None:
        """Set positioning mode to NO LIMITS."""
        if self.mc.info.register_exists(
            self.POSITIONING_OPTION_CODE_REGISTER, servo=self.servo, axis=self.axis
        ):
            self.mc.communication.set_register(
                self.POSITIONING_OPTION_CODE_REGISTER, 0, servo=self.servo, axis=self.axis
            )
        if self.mc.info.register_exists(
            self.MIN_POSITION_RANGE_LIMIT_REGISTER, servo=self.servo, axis=self.axis
        ):
            self.mc.communication.set_register(
                self.MIN_POSITION_RANGE_LIMIT_REGISTER, 0, servo=self.servo, axis=self.axis
            )
        if self.mc.info.register_exists(
            self.MAX_POSITION_RANGE_LIMIT_REGISTER, servo=self.servo, axis=self.axis
        ):
            self.mc.communication.set_register(
                self.MAX_POSITION_RANGE_LIMIT_REGISTER, 0, servo=self.servo, axis=self.axis
            )

    def __set_resolution_multiplier(self) -> None:
        """Set the resolution multiplier.

        Change multiplier using gear ratio if feedback to check is configured as position sensor
        (out of gear).
        """
        position_feedback_value = self.mc.configuration.get_position_feedback(
            servo=self.servo, axis=self.axis
        )
        velocity_feedback_value = self.mc.configuration.get_velocity_feedback(
            servo=self.servo, axis=self.axis
        )

        self.pos_vel_same_feedback = position_feedback_value == velocity_feedback_value
        if position_feedback_value == self.sensor:
            resolution_multiplier = self.mc.configuration.get_pos_to_vel_ratio(
                servo=self.servo, axis=self.axis
            )
            if not isinstance(resolution_multiplier, float):
                raise TypeError("Resolution multiplier has to be a float")
            self.resolution_multiplier = resolution_multiplier

    def teardown(self) -> None:
        self.logger.info("Disabling motor")
        self.mc.motion.motor_disable(servo=self.servo, axis=self.axis)

    @BaseTest.stoppable
    def wait_for_movement(self, timeout: float) -> None:
        timeout = time.time() + timeout
        while time.time() < timeout:
            time.sleep(0.1)
            if self.mc.errors.is_fault_active(servo=self.servo, axis=self.axis):
                self.show_error_message()

    @BaseTest.stoppable
    def get_current_position(self) -> float:
        position = self.mc.motion.get_actual_position(servo=self.servo, axis=self.axis)
        if not isinstance(position, int):
            raise TypeError("Actual position register must be an integer variable")
        current_position = position / self.resolution_multiplier
        return current_position

    @BaseTest.stoppable
    def current_ramp_up(self) -> None:
        rated_current = self.mc.communication.get_register(
            self.RATED_CURRENT_REGISTER, servo=self.servo, axis=self.axis
        )
        if not isinstance(rated_current, float):
            raise TypeError("Rated current has to be a float")
        nominal_current = self.mc.communication.get_register(
            self.MAXIMUM_CONTINUOUS_CURRENT_DRIVE_PROTECTION, servo=self.servo, axis=self.axis
        )
        if not isinstance(nominal_current, float):
            raise TypeError("Nominal current has to be a float")
        dict_currents = {
            "Rated motor current": rated_current,
            "Drive nominal current": nominal_current,
        }
        max_current = min(dict_currents.values())

        self.logger.debug(
            f"The maximum current is set by: {min(dict_currents, key=dict_currents.__getitem__)}"
        )
        # Increase current progressively
        self.logger.info(
            f"Increasing current to {self.PERCENTAGE_CURRENT_USED * 100}% "
            f"rated until one electrical cycle is completed"
        )

        target_current = self.PERCENTAGE_CURRENT_USED * max_current
        cycle_time = 2 / self.test_frequency

        self.mc.motion.current_quadrature_ramp(
            target_current, cycle_time, servo=self.servo, axis=self.axis
        )

    def first_movement_and_set_current(self) -> float:
        self.mc.motion.internal_generator_saw_tooth_move(
            1, 1, self.test_frequency, servo=self.servo, axis=self.axis
        )
        self.logger.info("Generator mode set to Saw tooth")
        self.logger.info("Generator frequency set to %s Hz", self.test_frequency)
        self.logger.info("Generator gain set to 1")
        self.logger.info("Generator offset set to 0")
        self.logger.info("Generator cycle number set to 1")
        self.current_ramp_up()
        self.wait_for_movement(self.TIME_BETWEEN_MOVEMENT)
        return self.get_current_position()

    @BaseTest.stoppable
    def internal_generator_move(self, polarity: Polarity) -> float:
        cycles = 1
        freq = self.test_frequency
        gain = 1 if polarity == self.Polarity.NORMAL else -1
        pol = 1 if polarity == self.Polarity.NORMAL else -1
        self.mc.motion.internal_generator_saw_tooth_move(
            pol, cycles, freq, servo=self.servo, axis=self.axis
        )
        self.logger.info("%s direction test", polarity.name)
        self.logger.info("Generator gain set to %s", gain)
        self.logger.info("Generator offset set to %s", polarity)
        self.logger.info("Generator Cycle number set to %s", cycles)
        self.logger.info("Wait until one electrical cycle is completed")
        self.wait_for_movement(cycles / freq)
        self.wait_for_movement(self.TIME_BETWEEN_MOVEMENT)
        position = self.get_current_position()
        self.logger.info("Actual position: %.0f", position)
        return position

    @BaseTest.stoppable
    def check_movement(self, position_displacement: float) -> None:
        self.logger.info("Detected forward displacement: %.0f", position_displacement)

        # Check the movement displacement
        if position_displacement == 0:
            error_movement_displacement = (
                "ERROR: No movement detected. "
                "Please, review your feedback "
                "configuration & wiring"
            )
            raise TestError(error_movement_displacement)

    def check_pos_vel_ratio(self) -> Optional[ResultType]:
        pos_vel_ratio = self.mc.configuration.get_pos_to_vel_ratio(servo=self.servo, axis=self.axis)
        if not isinstance(pos_vel_ratio, float):
            raise TypeError("Position to velocity sensor ratio value has to be a float")
        if self.pos_vel_same_feedback and not math.isclose(pos_vel_ratio, 1):
            return self.ResultType.POS_VEL_RATIO_ERROR
        if not self.pos_vel_same_feedback and math.isclose(pos_vel_ratio, 1):
            self.logger.warning(
                "Position and velocity feedbacks are different but"
                " the Position to velocity sensor ratio is 1."
            )
            return None
        else:
            return None

    @BaseTest.stoppable
    def loop(self) -> ResultType:
        self.logger.info("START OF THE TEST")
        check_pos_vel_output = self.check_pos_vel_ratio()
        if check_pos_vel_output is not None:
            return check_pos_vel_output
        try:
            self.mc.motion.motor_enable(servo=self.servo, axis=self.axis)
        except (ILTimeoutError, ILStateError, ILIOError) as e:
            raise TestError(f"An error occurred enabling motor. Reason: {e}")
        position_1 = self.first_movement_and_set_current()
        self.logger.info("Actual position: %.0f", position_1, axis=self.axis)
        position_2 = self.internal_generator_move(self.Polarity.NORMAL)
        position_displacement = position_2 - position_1
        self.check_movement(position_displacement)
        position_3 = self.internal_generator_move(self.Polarity.REVERSED)
        self.mc.motion.motor_disable(servo=self.servo, axis=self.axis)
        negative_displacement = position_3 - position_2
        self.logger.info("Detected reverse displacement: %.0f", negative_displacement)
        return self.generate_output(position_displacement, negative_displacement)

    def generate_output(
        self, position_displacement: float, negative_displacement: float
    ) -> ResultType:
        symmetry_check_result = self.check_symmetry(position_displacement, negative_displacement)
        if symmetry_check_result != self.ResultType.SUCCESS.value:
            return self.ResultType.SYMMETRY_ERROR
        resolution_check_result = self.check_resolution(position_displacement)
        if resolution_check_result != self.ResultType.SUCCESS.value:
            return self.ResultType.RESOLUTION_ERROR
        polarity = self.check_polarity(position_displacement)
        self.suggest_polarity(polarity)
        return self.ResultType.SUCCESS

    def get_result_msg(self, output: ResultType) -> str:
        return self.result_description[output]

    def get_result_severity(self, output: ResultType) -> SeverityLevel:
        if output < self.ResultType.SUCCESS:
            return SeverityLevel.FAIL
        else:
            return SeverityLevel.SUCCESS



import typing
from time import time
from functools import wraps
from typing import TypeVar, Callable, Any


class StopException(Exception):
    """Stop exception."""


T = typing.TypeVar("T")


class Stoppable:

    is_stopped = False

    @staticmethod
    def stoppable(fun: Callable[..., T]) -> Callable[..., T]:
        @wraps(fun)
        def wrapper(self, *args, **kwargs):  # type: ignore
            self.check_stop()
            return fun(self, *args, **kwargs)

        return wrapper

    def reset_stop(self) -> None:
        self.is_stopped = False

    def stop(self) -> None:
        self.is_stopped = True

    def check_stop(self) -> None:
        if self.is_stopped:
            raise StopException

    def stoppable_sleep(self, timeout: float) -> None:
        init_time = time()
        while init_time + timeout > time():
            self.check_stop()



from enum import IntEnum
from typing import TYPE_CHECKING

import ingenialogger

from ingeniamotion.wizard_tests.base_test import BaseTest, TestError
from ingeniamotion.enums import SensorType, FeedbackPolarity, OperationMode, SeverityLevel

if TYPE_CHECKING:
    from ingeniamotion.motion_controller import MotionController


class DCFeedbacksPolarityTest(BaseTest):
    MOVEMENT_ERROR_FACTOR = 0.05
    CURRENT_RAMP_TOTAL_TIME = 5
    CURRENT_RAMP_INTERVAL = 0.1
    OPERATION_MODE = OperationMode.CURRENT

    class ResultType(IntEnum):
        SUCCESS = 0

    result_description = {
        ResultType.SUCCESS: "Feedback polarity test pass successfully",
    }

    BACKUP_REGISTERS = [
        "CL_POS_FBK_SENSOR",
        "CL_VEL_FBK_SENSOR",
        "CL_AUX_FBK_SENSOR",
        "DRV_OP_CMD",
        "CL_CUR_Q_SET_POINT",
    ]

    feedback_resolution: int

    def __init__(self, mc: "MotionController", sensor: SensorType, servo: str, axis: int):
        if sensor == SensorType.HALLS:
            raise NotImplementedError("This test is not implemented for Hall sensor")
        super().__init__()
        self.mc = mc
        self.sensor = sensor
        self.servo = servo
        self.axis = axis
        self.BACKUP_REGISTERS.append(mc.configuration.get_feedback_polarity_register_uid(sensor))
        self.backup_registers_names = self.BACKUP_REGISTERS
        self.logger = ingenialogger.get_logger(__name__, axis=axis, drive=mc.servo_name(servo))

    @BaseTest.stoppable
    def setup(self) -> None:
        self.mc.motion.motor_disable(servo=self.servo, axis=self.axis)
        self.logger.info("Motor disable")
        self.feedback_resolution = self.mc.configuration.get_feedback_resolution(
            self.sensor, servo=self.servo, axis=self.axis
        )
        self.mc.configuration.set_feedback_polarity(
            FeedbackPolarity.NORMAL, self.sensor, servo=self.servo, axis=self.axis
        )
        self.logger.info(f"Set polarity to {FeedbackPolarity.NORMAL.name}")
        self.mc.motion.set_current_quadrature(0, servo=self.servo, axis=self.axis)
        self.logger.info("Set current to 0")
        self.mc.motion.set_operation_mode(self.OPERATION_MODE, servo=self.servo, axis=self.axis)
        self.logger.info(f"Set operation mode to {self.OPERATION_MODE.name}")
        self.mc.configuration.set_velocity_feedback(self.sensor, servo=self.servo, axis=self.axis)
        self.mc.configuration.set_position_feedback(self.sensor, servo=self.servo, axis=self.axis)
        if self.sensor == SensorType.BISSC2:
            self.mc.configuration.set_auxiliar_feedback(
                SensorType.ABS1, servo=self.servo, axis=self.axis
            )
            self.logger.info(
                f"Set velocity and position feedbacks to {self.sensor.name}"
                f" and axuiliar to {SensorType.ABS1.name}"
            )
        else:
            self.mc.configuration.set_auxiliar_feedback(
                self.sensor, servo=self.servo, axis=self.axis
            )
            self.logger.info(f"Set velocity, position and auxiliar feedbacks to {self.sensor.name}")

    @BaseTest.stoppable
    def increase_current_until_movement(self, initial_position: int, max_current: float) -> int:
        """Increase motor current until it moves

        Args:
            initial_position: initial position
            max_current: motor rated current

        Returns:
            returns final position

        Raises:
            TestError: No movement detected

        """
        for set_curr in self.mc.motion.ramp_generator(
            0, max_current, self.CURRENT_RAMP_TOTAL_TIME, self.CURRENT_RAMP_INTERVAL
        ):
            self.check_stop()
            current_pos = self.mc.motion.get_actual_position(servo=self.servo, axis=self.axis)
            if (
                abs(current_pos - initial_position)
                > self.feedback_resolution * self.MOVEMENT_ERROR_FACTOR
            ):
                self.mc.motion.motor_disable(self.servo, self.axis)
                return current_pos
            self.mc.motion.set_current_quadrature(set_curr, servo=self.servo, axis=self.axis)
            self.logger.info(f"Set current to {set_curr}")
        self.mc.motion.motor_disable(self.servo, self.axis)
        raise TestError(
            "ERROR: No movement detected. Please, review your feedback configuration & wiring"
        )

    @staticmethod
    def calculate_polarity(initial_position: int, final_position: int) -> FeedbackPolarity:
        """Calculate motor polarity

        Args:
            initial_position: initial position
            final_position: final position

        Returns:
            If final position is smaller than initial position return REVERSED, else NORMAL

        """
        if final_position - initial_position < 0:
            return FeedbackPolarity.REVERSED
        return FeedbackPolarity.NORMAL

    @BaseTest.stoppable
    def loop(self) -> ResultType:
        rated_current = self.mc.configuration.get_rated_current(servo=self.servo, axis=self.axis)
        max_current = self.mc.configuration.get_max_current(servo=self.servo, axis=self.axis)
        test_current = min(rated_current, max_current)
        self.mc.motion.motor_enable(servo=self.servo, axis=self.axis)
        self.logger.info("Motor enable")
        initial_position = self.mc.motion.get_actual_position(servo=self.servo, axis=self.axis)
        final_position = self.increase_current_until_movement(initial_position, test_current)
        polarity = self.calculate_polarity(initial_position, final_position)
        self.logger.info(f"Polarity found: {polarity.name}")
        polarity_uid = self.mc.configuration.get_feedback_polarity_register_uid(self.sensor)
        self.suggested_registers[polarity_uid] = polarity
        return self.ResultType.SUCCESS

    def teardown(self) -> None:
        self.mc.motion.motor_disable(servo=self.servo, axis=self.axis)
        self.logger.info("Motor disable")

    def get_result_msg(self, output: ResultType) -> str:
        description = self.result_description.get(output)
        if description is None:
            raise NotImplementedError("Result description not implemented")
        return description

    def get_result_severity(self, output: ResultType) -> SeverityLevel:
        if output == self.ResultType.SUCCESS:
            return SeverityLevel.SUCCESS
        return SeverityLevel.FAIL



from collections import deque
from typing import Dict
from unittest.mock import Mock

import pytest
from ingenialink import CAN_BAUDRATE, CAN_DEVICE
from ingenialink.exceptions import ILFirmwareLoadError
from ingenialink.pdo import RPDOMap, TPDOMap

from examples.change_baudrate import change_baudrate
from examples.change_node_id import change_node_id
from examples.commutation_test_encoders import main as main_commutation_test_encoders
from examples.connect_ecat_coe import connect_ethercat_coe
from examples.load_fw_canopen import load_firmware_canopen
from examples.load_save_config_register_changes import (
    main as main_load_save_config_register_changes,
)
from examples.load_save_configuration import main as main_load_save_configuration
from examples.pdo_poller_example import main as set_up_pdo_poller
from examples.position_ramp import main as main_position_ramp
from examples.process_data_object import main as main_process_data_object
from ingeniamotion import MotionController
from ingeniamotion.communication import Communication
from ingeniamotion.configuration import Configuration
from ingeniamotion.drive_tests import DriveTests
from ingeniamotion.enums import SeverityLevel
from ingeniamotion.information import Information
from ingeniamotion.motion import Motion
from ingeniamotion.pdo import PDONetworkManager, PDOPoller
from tests.conftest import connect_canopen, connect_eoe, connect_soem


@pytest.fixture
def setup_for_test_examples(motion_controller):
    mc, alias = motion_controller
    mc.communication.disconnect(alias)


@pytest.fixture
def teardown_for_test_examples(motion_controller, read_config, pytestconfig):
    yield
    mc, alias = motion_controller
    protocol = pytestconfig.getoption("--protocol")
    if protocol == "soem":
        connect_soem(mc, read_config, alias)
    elif protocol == "canopen":
        connect_canopen(mc, read_config, alias)
    else:
        connect_eoe(mc, read_config, alias)


from ingeniamotion.motion import Motion


@pytest.mark.eoe
def test_disturbance_example(read_config, script_runner):
    script_path = "examples/disturbance_example.py"
    ip_address = read_config["ip"]
    dictionary = read_config["dictionary"]
    result = script_runner.run(script_path, f"--ip={ip_address}", f"--dictionary_path={dictionary}")
    assert result.returncode == 0


@pytest.mark.usefixtures("setup_for_test_examples", "teardown_for_test_examples")
@pytest.mark.canopen
def test_canopen_example(read_config, script_runner):
    script_path = "examples/canopen_example.py"
    dictionary = read_config["dictionary"]
    node_id = read_config["node_id"]
    can_transceiver = read_config["device"]
    can_baudrate = read_config["baudrate"]
    can_channel = read_config["channel"]
    result = script_runner.run(
        script_path,
        f"--dictionary_path={dictionary}",
        f"--node_id={node_id}",
        f"--can_transceiver={can_transceiver}",
        f"--can_baudrate={can_baudrate}",
        f"--can_channel={can_channel}",
    )
    assert result.returncode == 0


@pytest.mark.eoe
def test_set_get_register_example(read_config, script_runner):
    script_path = "examples/set_get_register.py"
    ip_address = read_config["ip"]
    dictionary = read_config["dictionary"]
    result = script_runner.run(script_path, f"--ip={ip_address}", f"--dictionary_path={dictionary}")
    assert result.returncode == 0


@pytest.mark.eoe
def test_poller_example(read_config, script_runner):
    script_path = "examples/poller_example.py"
    ip_address = read_config["ip"]
    dictionary = read_config["dictionary"]
    result = script_runner.run(
        script_path, f"--ip={ip_address}", f"--dictionary_path={dictionary}", "--close"
    )
    assert result.returncode == 0


@pytest.mark.eoe
@pytest.mark.parametrize(
    "mode",
    ["velocity", "torque"],
)
def test_velocity_torque_ramp_example(read_config, script_runner, mocker, mode):
    script_path = "examples/velocity_torque_ramp.py"
    ip_address = read_config["ip"]
    dictionary = read_config["dictionary"]

    class MockMotion:
        def wait_for_velocity(self, *args, **kwargs):
            pass

        def set_operation_mode(self, *args, **kwargs):
            pass

        def set_velocity(self, *args, **kwargs):
            pass

        def motor_enable(*args, **kwargs):
            pass

        def motor_disable(*args, **kwargs):
            pass

        def set_current_quadrature(*args, **kwargs):
            pass

    mocker.patch.object(MotionController, "motion", MockMotion)
    mocker.patch("time.sleep", return_value=None)
    result = script_runner.run(
        script_path, mode, dictionary, f"-ip={ip_address}", f"-target_torque={0}"
    )
    assert result.returncode == 0


@pytest.mark.eoe
def test_monitoring_example(read_config, script_runner):
    script_path = "examples/monitoring_example.py"
    ip_address = read_config["ip"]
    dictionary = read_config["dictionary"]
    result = script_runner.run(
        script_path, f"--ip={ip_address}", f"--dictionary_path={dictionary}", "--close"
    )
    assert result.returncode == 0


@pytest.mark.eoe
def test_load_fw_ftp(read_config, script_runner, mocker):
    script_path = "examples/load_fw_ftp.py"
    ip_address = read_config["ip"]
    dictionary = read_config["dictionary"]
    fw_file = read_config["fw_file"]

    class MockCommunication:
        def boot_mode_and_load_firmware_ethernet(self, *args, **kwargs):
            pass

        def connect_servo_ethernet(self, *args, **kwargs):
            pass

    mocker.patch.object(MotionController, "communication", MockCommunication)
    result = script_runner.run(
        script_path,
        f"--dictionary_path={dictionary}",
        f"--ip={ip_address}",
        f"--firmware_file={fw_file}",
    )
    assert result.returncode == 0


@pytest.mark.soem
def test_load_fw_ecat(read_config, script_runner, mocker):
    script_path = "examples/load_fw_ecat.py"
    interface_index = read_config["index"]
    slave_id = read_config["slave"]
    fw_file = read_config["fw_file"]

    class MockCommunication:
        def load_firmware_ecat_interface_index(self, *args, **kwargs):
            pass

        def get_interface_name_list(*args, **kwargs):
            pass

    mocker.patch.object(MotionController, "communication", MockCommunication)
    result = script_runner.run(
        script_path,
        f"--interface_index={interface_index}",
        f"--slave_id={slave_id}",
        f"--firmware_file={fw_file}",
    )
    assert result.returncode == 0


@pytest.mark.eoe
@pytest.mark.parametrize(
    "feedback",
    ["HALLS", "QEI", "QEI2"],
)
def test_feedback_example(read_config, script_runner, mocker, feedback):
    script_path = "examples/feedback_test.py"
    ip_address = read_config["ip"]
    dictionary = read_config["dictionary"]

    class MockDriveTests:
        def digital_halls_test(*args, **kwargs):
            return {"result_message": SeverityLevel.SUCCESS}

        def incremental_encoder_1_test(*args, **kwargs):
            return {"result_message": SeverityLevel.SUCCESS}

        def incremental_encoder_2_test(*args, **kwargs):
            return {"result_message": SeverityLevel.SUCCESS}

    mocker.patch.object(MotionController, "tests", MockDriveTests)
    result = script_runner.run(script_path, feedback, dictionary, f"-ip={ip_address}")
    assert result.returncode == 0


@pytest.mark.eoe
def test_commutation_test_example(read_config, script_runner, mocker):
    script_path = "examples/commutation_test.py"
    ip_address = read_config["ip"]
    dictionary = read_config["dictionary"]

    class MockDriveTests:
        def commutation(*args, **kwargs):
            return {"result_message": SeverityLevel.SUCCESS}

    mocker.patch.object(MotionController, "tests", MockDriveTests)
    result = script_runner.run(script_path, dictionary, f"-ip={ip_address}")
    assert result.returncode == 0


@pytest.mark.eoe
@pytest.mark.parametrize(
    "override",
    ["disabled", "release", "enable"],
)
def test_brake_config_example(read_config, script_runner, mocker, override):
    script_path = "examples/brake_config.py"
    ip_address = read_config["ip"]
    dictionary = read_config["dictionary"]

    class MockConfiguration:
        def disable_brake_override(*args, **kwargs):
            pass

        def release_brake(*args, **kwargs):
            pass

        def enable_brake(*args, **kwargs):
            pass

    mocker.patch.object(MotionController, "configuration", MockConfiguration)
    result = script_runner.run(script_path, override, dictionary, f"-ip={ip_address}")
    assert result.returncode == 0


@pytest.mark.virtual
def test_can_bootloader_example_success(mocker, capsys):
    device = CAN_DEVICE.PCAN
    channel = 0
    baudrate = CAN_BAUDRATE.Baudrate_1M
    node_id = 32
    dictionary_path = "test_dictionary.xdf"
    fw_path = "test_fw.lfu"

    expected_node_list = [node_id]
    status_message = "Mock status message."
    progress_message = "100"

    class MockCommunication:
        def scan_servos_canopen(*args, **kwargs):
            return expected_node_list

        def connect_servo_canopen(*args, **kwargs):
            pass

        def load_firmware_canopen(*args, **kwargs):
            kwargs["status_callback"](status_message)
            kwargs["progress_callback"](progress_message)

        def disconnect(*args, **kwargs):
            pass

    mock_servos = {"default": "my_servo"}

    mocker.patch.object(MotionController, "communication", MockCommunication)
    mocker.patch.object(MotionController, "servos", mock_servos)
    load_firmware_canopen(device, channel, baudrate, node_id, dictionary_path, fw_path)

    captured_outputs = capsys.readouterr()
    all_outputs = captured_outputs.out.split("\n")
    assert all_outputs[0] == f"Found nodes: {expected_node_list}"
    assert all_outputs[1] == "Starts to establish a communication."
    assert all_outputs[2] == "Drive is connected."
    assert all_outputs[3] == "Starts to load the firmware."
    assert all_outputs[4] == f"Load firmware status: {status_message}"
    assert all_outputs[5] == f"Load firmware progress: {progress_message}%"
    assert all_outputs[6] == "Firmware is uploaded successfully."
    assert all_outputs[7] == "Drive is disconnected."


@pytest.mark.virtual
def test_can_bootloader_example_failed(mocker, capsys):
    device = CAN_DEVICE.PCAN
    channel = 0
    baudrate = CAN_BAUDRATE.Baudrate_1M
    node_id = 32
    dictionary_path = "test_dictionary.xdf"
    fw_path = "test_fw.lfu"

    expected_node_list = [node_id]
    fw_error_message = "An error occurs during the firmware updating."

    class MockCommunication:
        def scan_servos_canopen(*args, **kwargs):
            return expected_node_list

        def connect_servo_canopen(*args, **kwargs):
            pass

        def load_firmware_canopen(*args, **kwargs):
            raise ILFirmwareLoadError(fw_error_message)

        def disconnect(*args, **kwargs):
            pass

    mock_servos = {"default": "my_servo"}

    mocker.patch.object(MotionController, "communication", MockCommunication)
    mocker.patch.object(MotionController, "servos", mock_servos)
    load_firmware_canopen(device, channel, baudrate, node_id, dictionary_path, fw_path)

    captured_outputs = capsys.readouterr()
    all_outputs = captured_outputs.out.split("\n")
    assert all_outputs[0] == f"Found nodes: {expected_node_list}"
    assert all_outputs[1] == "Starts to establish a communication."
    assert all_outputs[2] == "Drive is connected."
    assert all_outputs[3] == "Starts to load the firmware."
    assert all_outputs[4] == f"Firmware loading failed: {fw_error_message}"
    assert all_outputs[5] == "Drive is disconnected."


@pytest.mark.virtual
def test_change_node_id_success(mocker, capsys):
    device = CAN_DEVICE.PCAN
    channel = 0
    baudrate = CAN_BAUDRATE.Baudrate_1M
    dictionary_path = "test_dictionary.xdf"

    node_id = 20
    test_new_node_id = 32

    mocker.patch.object(Communication, "connect_servo_canopen")
    mocker.patch.object(Communication, "disconnect")
    mocker.patch.object(Information, "get_node_id", side_effect=[node_id, test_new_node_id])
    mocker.patch.object(Configuration, "change_node_id")
    change_node_id(device, channel, node_id, baudrate, dictionary_path, test_new_node_id)

    captured_outputs = capsys.readouterr()
    all_outputs = captured_outputs.out.split("\n")
    assert all_outputs[1] == f"Drive is connected with {node_id} as a node ID."
    assert all_outputs[3] == "Node ID has been changed"
    assert all_outputs[6] == f"Now the drive is connected with {test_new_node_id} as a node ID."


@pytest.mark.virtual
def test_change_node_id_failed(mocker, capsys):
    device = CAN_DEVICE.PCAN
    channel = 0
    baudrate = CAN_BAUDRATE.Baudrate_1M
    dictionary_path = "test_dictionary.xdf"

    node_id = 20
    test_new_node_id = node_id

    mocker.patch.object(Communication, "connect_servo_canopen")
    mocker.patch.object(Communication, "disconnect")
    mocker.patch.object(Information, "get_node_id", side_effect=[node_id, test_new_node_id])
    mocker.patch.object(Configuration, "change_node_id")
    change_node_id(device, channel, node_id, baudrate, dictionary_path, test_new_node_id)

    captured_outputs = capsys.readouterr()
    all_outputs = captured_outputs.out.split("\n")
    assert all_outputs[1] == f"Drive is connected with {node_id} as a node ID."
    assert all_outputs[3] == f"This drive already has this node ID: {node_id}."


@pytest.mark.virtual
def test_change_baudrate_success(mocker, capsys):
    device = CAN_DEVICE.PCAN
    channel = 0
    baudrate = CAN_BAUDRATE.Baudrate_1M
    node_id = 32
    dictionary_path = "test_dictionary.xdf"
    test_new_baudrate = CAN_BAUDRATE.Baudrate_125K

    mocker.patch.object(Communication, "connect_servo_canopen")
    mocker.patch.object(Communication, "disconnect")
    mocker.patch.object(Information, "get_baudrate", side_effect=[baudrate])
    mocker.patch.object(Configuration, "change_baudrate")
    change_baudrate(device, channel, node_id, baudrate, dictionary_path, test_new_baudrate)

    captured_outputs = capsys.readouterr()
    all_outputs = captured_outputs.out.split("\n")
    assert all_outputs[0] == f"Drive is connected with {baudrate} baudrate."
    assert all_outputs[2] == f"Baudrate has been changed from {baudrate} to {test_new_baudrate}."
    assert (
        all_outputs[4]
        == f"Perform a power cycle and reconnect to the drive using the new baud rate: {test_new_baudrate}"
    )


@pytest.mark.virtual
def test_change_baudrate_failed(mocker, capsys):
    device = CAN_DEVICE.PCAN
    channel = 0
    baudrate = CAN_BAUDRATE.Baudrate_1M
    node_id = 32
    dictionary_path = "test_dictionary.xdf"
    test_new_baudrate = baudrate

    mocker.patch.object(Communication, "connect_servo_canopen")
    mocker.patch.object(Communication, "disconnect")
    mocker.patch.object(Information, "get_baudrate", side_effect=[baudrate])
    mocker.patch.object(Configuration, "change_baudrate")
    change_baudrate(device, channel, node_id, baudrate, dictionary_path, test_new_baudrate)

    captured_outputs = capsys.readouterr()
    all_outputs = captured_outputs.out.split("\n")
    assert all_outputs[0] == f"Drive is connected with {baudrate} baudrate."
    assert all_outputs[2] == f"This drive already has this baudrate: {baudrate}."


@pytest.mark.virtual
def test_ecat_coe_connection_example_success(mocker, capsys):
    interface_index = 2
    slave_id = 1
    dictionary_path = (
        "\\\\awe-srv-max-prd\\distext\\products\\CAP-NET\\firmware\\2.4.0\\cap-net-e_eoe_2.4.0.xdf"
    )
    expected_slave_list = [slave_id]
    expected_interfaces_name_list = ["Interface 1", "Interface 2", "Interface 3"]
    expected_real_name_interface = f"\\Device\\NPF_real_name_interface_{interface_index}"
    test_alias = "default"
    test_servos: Dict[str, str] = {}

    def scan_servos_ethercat(*args, **kwargs):
        return expected_slave_list

    def get_interface_name_list(*args, **kwargs):
        return expected_interfaces_name_list

    def get_ifname_by_index(*args, **kwargs):
        return expected_real_name_interface

    def connect_servo_ethercat(*args, **kwargs):
        test_servos[test_alias] = "my_servo"

    def disconnect(*args, **kwargs):
        test_servos.pop(test_alias)

    mocker.patch.object(Communication, "scan_servos_ethercat", scan_servos_ethercat)
    mocker.patch.object(Communication, "get_interface_name_list", get_interface_name_list)
    mocker.patch.object(Communication, "get_ifname_by_index", get_ifname_by_index)
    mocker.patch.object(MotionController, "servos", test_servos)
    mocker.patch.object(Communication, "connect_servo_ethercat", connect_servo_ethercat)
    mocker.patch.object(Communication, "disconnect", disconnect)
    connect_ethercat_coe(interface_index, slave_id, dictionary_path)
    captured_outputs = capsys.readouterr()
    all_outputs = captured_outputs.out.split("\n")
    assert all_outputs[0] == "List of interfaces:"
    assert all_outputs[1] == "0: Interface 1"
    assert all_outputs[2] == "1: Interface 2"
    assert all_outputs[3] == "2: Interface 3"
    assert all_outputs[4] == "Interface selected:"
    assert all_outputs[5] == f"- Index interface: {interface_index}"
    assert all_outputs[6] == f"- Interface identifier: {expected_real_name_interface}"
    assert all_outputs[7] == f"- Interface name: {expected_interfaces_name_list[interface_index]}"
    assert all_outputs[8] == f"Found slaves: {expected_slave_list}"
    assert all_outputs[9] == f"Drive is connected."
    assert all_outputs[10] == f"The drive has been disconnected."


@pytest.mark.virtual
def test_ecat_coe_connection_example_failed(mocker, capsys):
    interface_index = 2
    slave_id = 1
    dictionary_path = (
        "\\\\awe-srv-max-prd\\distext\\products\\CAP-NET\\firmware\\2.4.0\\cap-net-e_eoe_2.4.0.xdf"
    )
    expected_slave_list = []
    expected_interfaces_name_list = ["Interface 1", "Interface 2", "Interface 3"]
    expected_real_name_interface = f"\\Device\\NPF_real_name_interface_{interface_index}"
    test_servos: Dict[str, str] = {}

    def scan_servos_ethercat(*args, **kwargs):
        return expected_slave_list

    def get_interface_name_list(*args, **kwargs):
        return expected_interfaces_name_list

    def get_ifname_by_index(*args, **kwargs):
        return expected_real_name_interface

    mocker.patch.object(Communication, "scan_servos_ethercat", scan_servos_ethercat)
    mocker.patch.object(Communication, "get_interface_name_list", get_interface_name_list)
    mocker.patch.object(Communication, "get_ifname_by_index", get_ifname_by_index)
    mocker.patch.object(MotionController, "servos", test_servos)
    connect_ethercat_coe(interface_index, slave_id, dictionary_path)
    captured_outputs = capsys.readouterr()
    all_outputs = captured_outputs.out.split("\n")
    assert all_outputs[0] == "List of interfaces:"
    assert all_outputs[1] == "0: Interface 1"
    assert all_outputs[2] == "1: Interface 2"
    assert all_outputs[3] == "2: Interface 3"
    assert all_outputs[4] == "Interface selected:"
    assert all_outputs[5] == f"- Index interface: {interface_index}"
    assert all_outputs[6] == f"- Interface identifier: {expected_real_name_interface}"
    assert all_outputs[7] == f"- Interface name: {expected_interfaces_name_list[interface_index]}"
    assert (
        all_outputs[8]
        == f"No slave detected on interface: {expected_interfaces_name_list[interface_index]}"
    )


@pytest.mark.virtual
def test_ecat_coe_connection_example_connection_error(mocker, capsys):
    interface_index = 2
    slave_id = 1
    dictionary_path = (
        "\\\\awe-srv-max-prd\\distext\\products\\CAP-NET\\firmware\\2.4.0\\cap-net-e_eoe_2.4.0.xdf"
    )
    expected_interfaces_name_list = ["Interface 1", "Interface 2", "Interface 3"]
    expected_real_name_interface = f"\\Device\\NPF_real_name_interface_{interface_index}"

    def get_interface_name_list(*args, **kwargs):
        return expected_interfaces_name_list

    def get_ifname_by_index(*args, **kwargs):
        return expected_real_name_interface

    mocker.patch.object(Communication, "get_interface_name_list", get_interface_name_list)
    mocker.patch.object(Communication, "get_ifname_by_index", get_ifname_by_index)
    with pytest.raises(ConnectionError) as e:
        connect_ethercat_coe(interface_index, slave_id, dictionary_path)
    captured_outputs = capsys.readouterr()
    all_outputs = captured_outputs.out.split("\n")
    assert all_outputs[0] == "List of interfaces:"
    assert all_outputs[1] == "0: Interface 1"
    assert all_outputs[2] == "1: Interface 2"
    assert all_outputs[3] == "2: Interface 3"
    assert all_outputs[4] == "Interface selected:"
    assert all_outputs[5] == f"- Index interface: {interface_index}"
    assert all_outputs[6] == f"- Interface identifier: {expected_real_name_interface}"
    assert all_outputs[7] == f"- Interface name: {expected_interfaces_name_list[interface_index]}"
    assert e.value.args[0] == f"could not open interface {expected_real_name_interface}"


@pytest.mark.virtual
def test_pdo_poller_success(mocker):
    connect_servo_ethercat_interface_ip = mocker.patch.object(
        Communication, "connect_servo_ethercat_interface_ip"
    )
    disconnect = mocker.patch.object(Communication, "disconnect")
    mock_pdo_poller = PDOPoller(MotionController(), "mock_alias", 0.1, None, 100)
    create_poller = mocker.patch.object(
        PDONetworkManager, "create_poller", return_value=mock_pdo_poller
    )
    mock_poller_data = (deque([0.1, 0.2]), [deque([1, 2]), deque([0.0, 0.0])])
    data = mocker.patch.object(
        PDOPoller, "data", new_callable=mocker.PropertyMock, return_value=mock_poller_data
    )
    stop = mocker.patch.object(PDOPoller, "stop")

    set_up_pdo_poller()

    connect_servo_ethercat_interface_ip.assert_called_once()
    create_poller.assert_called_once()
    data.assert_called_once()
    stop.assert_called_once()
    disconnect.assert_called_once()


@pytest.mark.virtual
def test_load_save_configuration(mocker):
    connect_servo_ethercat_interface_index = mocker.patch.object(
        Communication, "connect_servo_ethercat_interface_index"
    )
    disconnect = mocker.patch.object(Communication, "disconnect")
    save_configuration = mocker.patch.object(Configuration, "save_configuration")
    load_configuration = mocker.patch.object(Configuration, "load_configuration")

    main_load_save_configuration()

    connect_servo_ethercat_interface_index.assert_called_once()
    save_configuration.assert_called_once()
    load_configuration.assert_called_once()
    disconnect.assert_called_once()


@pytest.mark.virtual
def test_load_save_configuration_register_changes_success(mocker, capsys):
    mocker.patch.object(Communication, "connect_servo_ethercat_interface_index")
    mocker.patch.object(Communication, "disconnect")
    mocker.patch.object(Configuration, "save_configuration")
    mocker.patch.object(Configuration, "load_configuration")
    test_velocities = [10.0, 10.0, 20.0]
    mocker.patch.object(Configuration, "get_max_velocity", side_effect=test_velocities)
    mocker.patch.object(Configuration, "set_max_velocity")

    main_load_save_config_register_changes()

    captured_outputs = capsys.readouterr()
    all_outputs = captured_outputs.out.split("\n")

    assert all_outputs[0] == "The initial configuration is saved."
    assert all_outputs[1] == "The configuration file is saved with the modification."
    assert (
        all_outputs[2]
        == f"Max. velocity register should be set to its initial value ({test_velocities[1]}). "
        f"Current value: {test_velocities[1]}"
    )
    assert (
        all_outputs[3]
        == f"Max. velocity register should now be set to the new value ({test_velocities[2]}). "
        f"Current value: {test_velocities[2]}"
    )


@pytest.mark.virtual
def test_load_save_configuration_register_changes_failed(mocker, capsys):
    mocker.patch.object(Communication, "connect_servo_ethercat_interface_index")
    mocker.patch.object(Communication, "disconnect")
    mocker.patch.object(Configuration, "save_configuration")
    mocker.patch.object(Configuration, "load_configuration")
    mocker.patch.object(Configuration, "get_max_velocity", return_value=20.0)
    mocker.patch.object(Configuration, "set_max_velocity")

    main_load_save_config_register_changes()

    captured_outputs = capsys.readouterr()
    all_outputs = captured_outputs.out.split("\n")

    assert all_outputs[0] == "The initial configuration is saved."
    assert all_outputs[1] == "This max. velocity value is already set."


@pytest.mark.virtual
def test_process_data_object(mocker):
    connect_servo_ethercat_interface_ip = mocker.patch.object(
        Communication, "connect_servo_ethercat_interface_ip"
    )
    disconnect = mocker.patch.object(Communication, "disconnect")
    motor_enable = mocker.patch.object(Motion, "motor_enable")
    motor_disable = mocker.patch.object(Motion, "motor_disable")
    create_pdo_item = mocker.patch.object(PDONetworkManager, "create_pdo_item")
    create_pdo_maps = mocker.patch.object(
        PDONetworkManager, "create_pdo_maps", return_value=(RPDOMap(), TPDOMap)
    )
    set_pdo_maps_to_slave = mocker.patch.object(PDONetworkManager, "set_pdo_maps_to_slave")
    start_pdos = mocker.patch.object(PDONetworkManager, "start_pdos")
    stop_pdos = mocker.patch.object(PDONetworkManager, "stop_pdos")
    mocker.patch.object(Motion, "get_actual_position")
    subscribe_to_receive_process_data = mocker.patch.object(
        PDONetworkManager, "subscribe_to_receive_process_data"
    )
    subscribe_to_send_process_data = mocker.patch.object(
        PDONetworkManager, "subscribe_to_send_process_data"
    )

    mocks_to_attach = {
        "connect_servo_ethercat_interface_ip": connect_servo_ethercat_interface_ip,
        "motor_enable": motor_enable,
        "create_pdo_item": create_pdo_item,
        "create_pdo_maps": create_pdo_maps,
        "set_pdo_maps_to_slave": set_pdo_maps_to_slave,
        "subscribe_to_receive_process_data": subscribe_to_receive_process_data,
        "subscribe_to_send_process_data": subscribe_to_send_process_data,
        "start_pdos": start_pdos,
        "stop_pdos": stop_pdos,
        "motor_disable": motor_disable,
        "disconnect": disconnect,
    }
    order_mock = Mock()
    for mock_name, mock in mocks_to_attach.items():
        order_mock.attach_mock(mock, f"{mock_name}")

    assert order_mock.method_calls == []

    main_process_data_object()

    expected_order_execution = [
        "connect_servo_ethercat_interface_ip",
        "motor_enable",
        "create_pdo_item",
        "create_pdo_item",
        "create_pdo_maps",
        "subscribe_to_receive_process_data",
        "subscribe_to_send_process_data",
        "set_pdo_maps_to_slave",
        "start_pdos",
        "stop_pdos",
        "motor_disable",
        "disconnect",
    ]
    for current_function, expected_function_name in enumerate(expected_order_execution):
        assert order_mock.method_calls[current_function][0] == expected_function_name


@pytest.mark.virtual
def test_commutation_test(mocker):
    connect_servo_ethercat_interface_ip = mocker.patch.object(
        Communication, "connect_servo_ethercat_interface_ip"
    )
    disconnect = mocker.patch.object(Communication, "disconnect")
    set_auxiliar_feedback = mocker.patch.object(Configuration, "set_auxiliar_feedback")
    set_commutation_feedback = mocker.patch.object(Configuration, "set_commutation_feedback")
    set_position_feedback = mocker.patch.object(Configuration, "set_position_feedback")
    set_velocity_feedback = mocker.patch.object(Configuration, "set_velocity_feedback")
    set_reference_feedback = mocker.patch.object(Configuration, "set_reference_feedback")
    commutation_test = mocker.patch.object(
        DriveTests,
        "commutation",
        return_value={
            "result_message": "Commutation is called",
            "result_severity": SeverityLevel.SUCCESS,
        },
    )

    main_commutation_test_encoders()

    connect_servo_ethercat_interface_ip.assert_called_once()
    set_auxiliar_feedback.assert_called_once()
    set_commutation_feedback.assert_called_once()
    set_position_feedback.assert_called_once()
    set_velocity_feedback.assert_called_once()
    set_reference_feedback.assert_called_once()
    commutation_test.assert_called_once()
    disconnect.assert_called_once()


@pytest.mark.virtual
def test_position_ramp(mocker, capsys):
    connect_servo_ethercat_interface_ip = mocker.patch.object(
        Communication, "connect_servo_ethercat_interface_ip"
    )
    disconnect = mocker.patch.object(Communication, "disconnect")
    set_max_profile_velocity = mocker.patch.object(Configuration, "set_max_profile_velocity")
    set_max_profile_acceleration = mocker.patch.object(
        Configuration, "set_max_profile_acceleration"
    )
    set_max_profile_deceleration = mocker.patch.object(
        Configuration, "set_max_profile_deceleration"
    )
    motor_enable = mocker.patch.object(Motion, "motor_enable")
    motor_disable = mocker.patch.object(Motion, "motor_disable")
    set_operation_mode = mocker.patch.object(Motion, "set_operation_mode")
    move_to_position = mocker.patch.object(Motion, "move_to_position")
    test_actual_values = [1500, 3000, 0]
    get_actual_position = mocker.patch.object(
        Motion, "get_actual_position", side_effect=test_actual_values
    )

    main_position_ramp()

    connect_servo_ethercat_interface_ip.assert_called_once()
    set_operation_mode.assert_called_once()
    set_max_profile_acceleration.assert_called_once()
    set_max_profile_deceleration.assert_called_once()
    set_max_profile_velocity.assert_called_once()
    motor_enable.assert_called_once()
    for current_target in test_actual_values:
        move_to_position.assert_any_call(current_target, blocking=True, timeout=2.0)
    assert move_to_position.call_count == len(test_actual_values)
    assert get_actual_position.call_count == len(test_actual_values)
    motor_disable.assert_called_once()
    disconnect.assert_called_once()

    captured_outputs = capsys.readouterr()
    all_outputs = captured_outputs.out.split("\n")
    for i, expected_actual_value in enumerate(test_actual_values):
        assert all_outputs[i] == f"Actual position: {expected_actual_value}"



import os

import pytest
from ingenialink.ethercat.servo import EthercatServo

from ingeniamotion.enums import (
    CommutationMode,
    FilterType,
    FilterNumber,
    FilterSignal,
    CAN_BAUDRATE,
)
from ingeniamotion.exceptions import IMException
from ingeniamotion.configuration import TYPE_SUBNODES

BRAKE_OVERRIDE_REGISTER = "MOT_BRAKE_OVERRIDE"
POSITION_SET_POINT_REGISTER = "CL_POS_SET_POINT_VALUE"
PROFILE_MAX_ACCELERATION_REGISTER = "PROF_MAX_ACC"
PROFILE_MAX_DECELERATION_REGISTER = "PROF_MAX_DEC"
PROFILE_MAX_VELOCITY_REGISTER = "PROF_MAX_VEL"
MAX_VELOCITY_REGISTER = "CL_VEL_REF_MAX"
POWER_STAGE_FREQUENCY_SELECTION_REGISTER = "DRV_PS_FREQ_SELECTION"
POSITION_AND_VELOCITY_LOOP_RATE_REGISTER = "DRV_POS_VEL_RATE"
CURRENT_LOOP_RATE_REGISTER = "CL_CUR_FREQ"
STATUS_WORD_REGISTER = "DRV_STATE_STATUS"
PHASING_MODE_REGISTER = "COMMU_PHASING_MODE"
GENERATOR_MODE_REGISTER = "FBK_GEN_MODE"
MOTOR_POLE_PAIRS_REGISTER = "MOT_PAIR_POLES"
STO_STATUS_REGISTER = "DRV_PROT_STO_STATUS"
VELOCITY_LOOP_KP_REGISTER = "CL_VEL_PID_KP"
VELOCITY_LOOP_KI_REGISTER = "CL_VEL_PID_KI"
VELOCITY_LOOP_KD_REGISTER = "CL_VEL_PID_KD"
POSITION_LOOP_KP_REGISTER = "CL_POS_PID_KP"
POSITION_LOOP_KI_REGISTER = "CL_POS_PID_KI"
POSITION_LOOP_KD_REGISTER = "CL_POS_PID_KD"
RATED_CURRENT_REGISTER = "MOT_RATED_CURRENT"
MAX_CURRENT_REGISTER = "CL_CUR_REF_MAX"
COMMUTATION_MODE_REGISTER = "MOT_COMMU_MOD"
BUS_VOLTAGE_REGISTER = "DRV_PROT_VBUS_VALUE"
POSITION_TO_VELOCITY_RATIO_REGISTER = "PROF_POS_VEL_RATIO"
FILTER_TYPE_REGISTER = "CL_{}_FILTER{}_TYPE"
FILTER_FREQ_REGISTER = "CL_{}_FILTER{}_FREQ"
FILTER_Q_REGISTER = "CL_{}_FILTER{}_Q"
FILTER_GAIN_REGISTER = "CL_{}_FILTER{}_GAIN"


@pytest.fixture
def teardown_brake_override(motion_controller):
    yield
    mc, alias = motion_controller
    mc.configuration.default_brake(servo=alias)


@pytest.mark.virtual
@pytest.mark.smoke
def test_release_brake(motion_controller, teardown_brake_override):
    mc, alias = motion_controller
    mc.configuration.release_brake(servo=alias)
    assert (
        mc.communication.get_register(BRAKE_OVERRIDE_REGISTER, servo=alias, axis=1)
        == mc.configuration.BrakeOverride.RELEASE_BRAKE
    )


@pytest.mark.virtual
@pytest.mark.smoke
def test_enable_brake(motion_controller, teardown_brake_override):
    mc, alias = motion_controller
    mc.configuration.enable_brake(servo=alias)
    assert (
        mc.communication.get_register(BRAKE_OVERRIDE_REGISTER, servo=alias, axis=1)
        == mc.configuration.BrakeOverride.ENABLE_BRAKE
    )


@pytest.mark.virtual
@pytest.mark.smoke
def test_disable_brake_override(motion_controller, teardown_brake_override):
    mc, alias = motion_controller
    mc.configuration.disable_brake_override(servo=alias)
    assert (
        mc.communication.get_register(BRAKE_OVERRIDE_REGISTER, servo=alias, axis=1)
        == mc.configuration.BrakeOverride.OVERRIDE_DISABLED
    )


@pytest.fixture
def remove_file_if_exist():
    yield
    file_path = "test_file.xcf"
    if os.path.isfile(file_path):
        os.remove(file_path)


@pytest.mark.smoke
@pytest.mark.usefixtures("remove_file_if_exist")
def test_save_configuration_and_load_configuration(motion_controller):
    file_path = "test_file.xcf"
    mc, alias = motion_controller
    old_value = mc.communication.get_register(PROFILE_MAX_VELOCITY_REGISTER, servo=alias)
    mc.communication.set_register(PROFILE_MAX_VELOCITY_REGISTER, 10, servo=alias)
    mc.configuration.save_configuration(file_path, servo=alias)
    assert os.path.isfile(file_path)
    mc.communication.set_register(PROFILE_MAX_VELOCITY_REGISTER, 20, servo=alias)
    mc.configuration.load_configuration(file_path, servo=alias)
    assert mc.communication.get_register(PROFILE_MAX_VELOCITY_REGISTER, servo=alias) == 10
    mc.communication.set_register(PROFILE_MAX_VELOCITY_REGISTER, old_value, servo=alias)


@pytest.mark.usefixtures("remove_file_if_exist")
@pytest.mark.canopen
@pytest.mark.eoe
@pytest.mark.smoke
def test_save_configuration_and_load_configuration_nvm_none(motion_controller):
    file_path = "test_file.xcf"
    mc, alias = motion_controller
    old_value = mc.communication.get_register(POSITION_SET_POINT_REGISTER, servo=alias)
    mc.communication.set_register(POSITION_SET_POINT_REGISTER, 10, servo=alias)
    mc.configuration.save_configuration(file_path, servo=alias)
    assert os.path.isfile(file_path)
    mc.communication.set_register(POSITION_SET_POINT_REGISTER, 20, servo=alias)
    mc.configuration.load_configuration(file_path, servo=alias)
    assert mc.communication.get_register(POSITION_SET_POINT_REGISTER, servo=alias) == 20
    mc.communication.set_register(POSITION_SET_POINT_REGISTER, old_value, servo=alias)


@pytest.mark.virtual
@pytest.mark.smoke
def test_set_profiler_exception(motion_controller):
    mc, alias = motion_controller

    with pytest.raises(TypeError):
        mc.configuration.set_profiler(None, None, None, servo=alias)


@pytest.mark.virtual
@pytest.mark.smoke
@pytest.mark.parametrize(
    "acceleration, deceleration, velocity",
    [(0, 0, 0), (15, 20, 25), (1, None, None), (None, 1, None), (None, None, 1)],
)
def test_set_profiler(motion_controller, acceleration, deceleration, velocity):
    mc, alias = motion_controller
    register_dict = {
        "acc": PROFILE_MAX_ACCELERATION_REGISTER,
        "dec": PROFILE_MAX_DECELERATION_REGISTER,
        "vel": PROFILE_MAX_VELOCITY_REGISTER,
    }
    expected_values = {"acc": acceleration, "dec": deceleration, "vel": velocity}
    for key in [key for key, value in expected_values.items() if value is None]:
        expected_value = mc.communication.get_register(register_dict[key], servo=alias)
        expected_values[key] = expected_value
    mc.configuration.set_profiler(acceleration, deceleration, velocity, servo=alias)
    for key, value in expected_values.items():
        actual_value = mc.communication.get_register(register_dict[key], servo=alias)
        assert pytest.approx(expected_values[key]) == actual_value


@pytest.mark.virtual
@pytest.mark.smoke
@pytest.mark.parametrize("acceleration", [0, 10, 25])
def test_set_max_acceleration(motion_controller, acceleration):
    mc, alias = motion_controller
    mc.configuration.set_max_acceleration(acceleration, servo=alias)
    output_value = mc.communication.get_register(PROFILE_MAX_ACCELERATION_REGISTER, servo=alias)
    assert pytest.approx(acceleration) == output_value


@pytest.mark.virtual
@pytest.mark.smoke
@pytest.mark.parametrize("acceleration", [0, 10, 25])
def test_set_max_profile_acceleration(motion_controller, acceleration):
    mc, alias = motion_controller
    mc.configuration.set_max_profile_acceleration(acceleration, servo=alias)
    output_value = mc.communication.get_register(PROFILE_MAX_ACCELERATION_REGISTER, servo=alias)
    assert pytest.approx(acceleration) == output_value


@pytest.mark.virtual
@pytest.mark.smoke
@pytest.mark.parametrize("deceleration", [0, 10, 25])
def test_set_max_deceleration(motion_controller, deceleration):
    mc, alias = motion_controller
    mc.configuration.set_max_profile_deceleration(deceleration, servo=alias)
    output_value = mc.communication.get_register(PROFILE_MAX_DECELERATION_REGISTER, servo=alias)
    assert pytest.approx(output_value) == deceleration


@pytest.mark.virtual
@pytest.mark.smoke
@pytest.mark.parametrize("velocity", [0, 10, 25])
def test_set_max_velocity(motion_controller, velocity):
    mc, alias = motion_controller
    mc.configuration.set_max_velocity(velocity, servo=alias)
    output_value = mc.communication.get_register(MAX_VELOCITY_REGISTER, servo=alias)
    assert pytest.approx(velocity) == output_value


@pytest.mark.virtual
@pytest.mark.smoke
@pytest.mark.parametrize("velocity", [0, 10, 25])
def test_set_max_profile_velocity(motion_controller, velocity):
    mc, alias = motion_controller
    mc.configuration.set_max_profile_velocity(velocity, servo=alias)
    output_value = mc.communication.get_register(PROFILE_MAX_VELOCITY_REGISTER, servo=alias)
    assert pytest.approx(velocity) == output_value


@pytest.mark.virtual
@pytest.mark.smoke
def test_get_position_and_velocity_loop_rate(motion_controller):
    mc, alias = motion_controller
    test_value = mc.configuration.get_position_and_velocity_loop_rate(servo=alias)
    reg_value = mc.communication.get_register(POSITION_AND_VELOCITY_LOOP_RATE_REGISTER, servo=alias)
    assert test_value == reg_value


@pytest.mark.virtual
@pytest.mark.smoke
def test_get_current_loop_rate(motion_controller):
    mc, alias = motion_controller
    test_value = mc.configuration.get_current_loop_rate(servo=alias)
    reg_value = mc.communication.get_register(CURRENT_LOOP_RATE_REGISTER, servo=alias)
    assert test_value == reg_value


@pytest.mark.virtual
@pytest.mark.smoke
def test_get_power_stage_frequency(motion_controller):
    mc, alias = motion_controller
    mc.configuration.get_power_stage_frequency(servo=alias)


@pytest.mark.virtual
@pytest.mark.smoke
def test_get_power_stage_frequency_raw(motion_controller):
    mc, alias = motion_controller
    test_value = mc.configuration.get_power_stage_frequency(servo=alias, raw=True)
    pow_stg_freq = mc.communication.get_register(
        POWER_STAGE_FREQUENCY_SELECTION_REGISTER, servo=alias
    )
    assert test_value == pow_stg_freq


@pytest.mark.virtual
@pytest.mark.smoke
def test_get_power_stage_frequency_enum(motion_controller):
    mc, alias = motion_controller
    mc.configuration.get_power_stage_frequency_enum(servo=alias)


@pytest.mark.virtual
@pytest.mark.smoke
@pytest.mark.parametrize("input_value", [0, 1, 2, 3])
def test_set_power_stage_frequency(motion_controller_teardown, input_value):
    input_value = 0
    mc, alias = motion_controller_teardown
    mc.configuration.set_power_stage_frequency(input_value, servo=alias)
    output_value = mc.communication.get_register(
        POWER_STAGE_FREQUENCY_SELECTION_REGISTER, servo=alias
    )
    assert pytest.approx(input_value) == output_value


@pytest.mark.virtual
@pytest.mark.smoke
def test_get_power_stage_frequency_exception(mocker, motion_controller):
    mc, alias = motion_controller
    mocker.patch("ingeniamotion.communication.Communication.get_register", return_value=5)
    with pytest.raises(ValueError):
        mc.configuration.get_power_stage_frequency(servo=alias)


@pytest.mark.virtual
@pytest.mark.smoke
def test_get_status_word(motion_controller):
    mc, alias = motion_controller
    test_value = mc.configuration.get_status_word(servo=alias)
    reg_value = mc.communication.get_register(STATUS_WORD_REGISTER, servo=alias)
    assert test_value == reg_value


@pytest.mark.smoke
def test_is_motor_enabled_1(motion_controller):
    mc, alias = motion_controller
    mc.motion.motor_disable(alias)
    assert not mc.configuration.is_motor_enabled(servo=alias)
    mc.motion.motor_enable(servo=alias)
    assert mc.configuration.is_motor_enabled(servo=alias)
    mc.motion.motor_disable(servo=alias)
    assert not mc.configuration.is_motor_enabled(servo=alias)


@pytest.mark.virtual
@pytest.mark.smoke
@pytest.mark.parametrize(
    "status_word_value, expected_result",
    [
        (0xF29C, True),
        (0xF440, False),
        (0xD1A7, True),
        (0x86D7, True),
        (0x2A43, False),
        (0x33E6, True),
    ],
)
def test_is_motor_enabled_2(mocker, motion_controller, status_word_value, expected_result):
    mc, alias = motion_controller
    mocker.patch(
        "ingeniamotion.configuration.Configuration.get_status_word", return_value=status_word_value
    )
    test_value = mc.configuration.is_motor_enabled(servo=alias)
    assert test_value == expected_result


@pytest.mark.virtual
@pytest.mark.smoke
@pytest.mark.parametrize(
    "status_word_value, expected_result",
    [
        (0xF29C, True),
        (0xF440, True),
        (0xD1A7, True),
        (0x86D7, False),
        (0x2A43, False),
        (0x33E6, False),
    ],
)
def test_is_commutation_feedback_aligned(
    mocker, motion_controller, status_word_value, expected_result
):
    mc, alias = motion_controller
    mocker.patch(
        "ingeniamotion.configuration.Configuration.get_status_word", return_value=status_word_value
    )
    test_value = mc.configuration.is_commutation_feedback_aligned(servo=alias)
    assert test_value == expected_result


@pytest.mark.virtual
@pytest.mark.smoke
def test_set_phasing_mode(motion_controller):
    input_value = 0
    mc, alias = motion_controller
    mc.configuration.set_phasing_mode(input_value, servo=alias)
    output_value = mc.communication.get_register(PHASING_MODE_REGISTER, servo=alias)
    assert pytest.approx(input_value) == output_value


@pytest.mark.virtual
@pytest.mark.smoke
def test_get_phasing_mode(motion_controller):
    mc, alias = motion_controller
    test_value = mc.configuration.get_phasing_mode(servo=alias)
    reg_value = mc.communication.get_register(PHASING_MODE_REGISTER, servo=alias)
    assert test_value == reg_value


@pytest.mark.virtual
@pytest.mark.smoke
def test_set_generator_mode(motion_controller):
    input_value = 0
    mc, alias = motion_controller
    mc.configuration.set_generator_mode(input_value, servo=alias)
    output_value = mc.communication.get_register(GENERATOR_MODE_REGISTER, servo=alias)
    assert pytest.approx(input_value) == output_value


@pytest.mark.virtual
@pytest.mark.smoke
def test_set_motor_pair_poles(motion_controller_teardown):
    input_value = 0
    mc, alias = motion_controller_teardown
    mc.configuration.set_motor_pair_poles(input_value, servo=alias)
    output_value = mc.communication.get_register(MOTOR_POLE_PAIRS_REGISTER, servo=alias)
    assert pytest.approx(input_value) == output_value


@pytest.mark.virtual
@pytest.mark.smoke
def test_get_motor_pair_poles(motion_controller):
    mc, alias = motion_controller
    test_value = mc.configuration.get_motor_pair_poles(servo=alias)
    reg_value = mc.communication.get_register(MOTOR_POLE_PAIRS_REGISTER, servo=alias)
    assert test_value == reg_value


@pytest.mark.virtual
@pytest.mark.smoke
def test_get_sto_status(motion_controller):
    mc, alias = motion_controller
    test_value = mc.configuration.get_sto_status(servo=alias)
    reg_value = mc.communication.get_register(STO_STATUS_REGISTER, servo=alias)
    assert test_value == reg_value


def patch_get_sto_status(mocker, value):
    mocker.patch("ingeniamotion.configuration.Configuration.get_sto_status", return_value=value)


@pytest.mark.virtual
@pytest.mark.smoke
@pytest.mark.parametrize(
    "sto_status_value, expected_result",
    [
        (0x4843, 1),
        (0xF567, 1),
        (0xFFFF, 1),
        (0x0000, 0),
        (0x4766, 0),
        (0xF6A4, 0),
    ],
)
def test_is_sto1_active(mocker, motion_controller, sto_status_value, expected_result):
    mc, alias = motion_controller
    patch_get_sto_status(mocker, sto_status_value)
    value = mc.configuration.is_sto1_active(servo=alias)
    assert value == expected_result


@pytest.mark.virtual
@pytest.mark.smoke
@pytest.mark.parametrize(
    "sto_status_value, expected_result",
    [(0xA187, 1), (0x31BA, 1), (0xD7DD, 0), (0xFB8, 0), (0xA8DE, 1), (0x99A5, 0)],
)
def test_is_sto2_active(mocker, motion_controller, sto_status_value, expected_result):
    mc, alias = motion_controller
    patch_get_sto_status(mocker, sto_status_value)
    value = mc.configuration.is_sto2_active(servo=alias)
    assert value == expected_result


@pytest.mark.virtual
@pytest.mark.smoke
@pytest.mark.parametrize(
    "sto_status_value, expected_result",
    [(0xFAC4, 1), (0x1AE1, 0), (0xD9CA, 0), (0xEE94, 1), (0xAE9F, 1), (0x478B, 0)],
)
def test_check_sto_power_supply(mocker, motion_controller, sto_status_value, expected_result):
    mc, alias = motion_controller
    patch_get_sto_status(mocker, sto_status_value)
    value = mc.configuration.check_sto_power_supply(servo=alias)
    assert value == expected_result


@pytest.mark.smoke
@pytest.mark.parametrize(
    "sto_status_value, expected_result",
    [(0x1BAF, 1), (0xD363, 0), (0xAD9D, 1), (0x8D14, 0), (0x9AEE, 1), (0x94A7, 0)],
)
def test_check_sto_abnormal_fault(mocker, motion_controller, sto_status_value, expected_result):
    mc, alias = motion_controller
    patch_get_sto_status(mocker, sto_status_value)
    value = mc.configuration.check_sto_abnormal_fault(servo=alias)
    assert value == expected_result


@pytest.mark.virtual
@pytest.mark.smoke
@pytest.mark.parametrize(
    "sto_status_value, expected_result",
    [(0xF29C, 1), (0xF440, 0), (0xD1A7, 0), (0x86D7, 1), (0x2A43, 0), (0x33E6, 0)],
)
def test_get_sto_report_bit(mocker, motion_controller, sto_status_value, expected_result):
    mc, alias = motion_controller
    patch_get_sto_status(mocker, sto_status_value)
    value = mc.configuration.get_sto_report_bit(servo=alias)
    assert value == expected_result


@pytest.mark.virtual
@pytest.mark.smoke
@pytest.mark.parametrize(
    "sto_status_value, expected_result", [(0x13A0, False), (0x7648, False), (0x4, True)]
)
def test_is_sto_active(mocker, motion_controller, sto_status_value, expected_result):
    mc, alias = motion_controller
    patch_get_sto_status(mocker, sto_status_value)
    value = mc.configuration.is_sto_active(servo=alias)
    assert value == expected_result


@pytest.mark.virtual
@pytest.mark.smoke
@pytest.mark.parametrize(
    "sto_status_value, expected_result", [(0xC18A, False), (0x742C, False), (0x17, True)]
)
def test_is_sto_inactive(mocker, motion_controller, sto_status_value, expected_result):
    mc, alias = motion_controller
    patch_get_sto_status(mocker, sto_status_value)
    value = mc.configuration.is_sto_inactive(servo=alias)
    assert value == expected_result


@pytest.mark.virtual
@pytest.mark.smoke
@pytest.mark.parametrize(
    "sto_status_value, expected_result", [(0x1BF3, False), (0x6B7, False), (0x1F, True)]
)
def test_is_sto_abnormal_latched(mocker, motion_controller, sto_status_value, expected_result):
    mc, alias = motion_controller
    patch_get_sto_status(mocker, sto_status_value)
    value = mc.configuration.is_sto_abnormal_latched(servo=alias)
    assert value == expected_result


@pytest.mark.smoke
def test_store_configuration(motion_controller):
    mc, alias = motion_controller
    mc.configuration.store_configuration(servo=alias)


@pytest.mark.smoke
def test_restore_configuration(motion_controller):
    mc, alias = motion_controller
    mc.configuration.restore_configuration(servo=alias)


@pytest.mark.virtual
def test_get_drive_info_coco_moco(motion_controller):
    expected_product_codes = [123456, 123456]
    expected_revision_numbers = [654321, 654321]
    expected_firmware_versions = ["0.1.0", "0.1.0"]
    expected_serial_numbers = [123456789, 123456789]

    mc, alias = motion_controller
    prod_codes, rev_nums, fw_vers, ser_nums = mc.configuration.get_drive_info_coco_moco(alias)

    assert prod_codes == expected_product_codes
    assert rev_nums == expected_revision_numbers
    assert fw_vers == expected_firmware_versions
    assert ser_nums == expected_serial_numbers


@pytest.mark.virtual
def test_get_product_code(motion_controller):
    expected_product_code_0 = 123456
    expected_product_code_1 = 123456

    mc, alias = motion_controller
    product_code_0 = mc.configuration.get_product_code(alias, 0)
    product_code_1 = mc.configuration.get_product_code(alias, 1)

    assert product_code_0 == expected_product_code_0
    assert product_code_1 == expected_product_code_1


@pytest.mark.virtual
def test_get_revision_number(motion_controller):
    expected_revision_number_0 = 654321
    expected_revision_number_1 = 654321

    mc, alias = motion_controller
    revision_number_0 = mc.configuration.get_revision_number(alias, 0)
    revision_number_1 = mc.configuration.get_revision_number(alias, 1)

    assert revision_number_0 == expected_revision_number_0
    assert revision_number_1 == expected_revision_number_1


@pytest.mark.virtual
def test_get_serial_number(motion_controller):
    expected_serial_number_0 = 123456789
    expected_serial_number_1 = 123456789

    mc, alias = motion_controller
    serial_number_0 = mc.configuration.get_serial_number(alias, 0)
    serial_number_1 = mc.configuration.get_serial_number(alias, 1)

    assert serial_number_0 == expected_serial_number_0
    assert serial_number_1 == expected_serial_number_1


@pytest.mark.virtual
def test_get_fw_version(motion_controller):
    expected_fw_version_0 = "0.1.0"
    expected_fw_version_1 = "0.1.0"

    mc, alias = motion_controller
    firmware_version_0 = mc.configuration.get_fw_version(alias, 0)
    firmware_version_1 = mc.configuration.get_fw_version(alias, 1)

    assert firmware_version_0 == expected_fw_version_0
    assert firmware_version_1 == expected_fw_version_1


@pytest.mark.virtual
def test_change_baudrate_exception(motion_controller):
    mc, alias = motion_controller
    with pytest.raises(ValueError):
        mc.configuration.change_baudrate(CAN_BAUDRATE.Baudrate_1M, alias)


@pytest.mark.virtual
def test_get_vendor_id(motion_controller):
    expected_vendor_id_0 = 987654321
    expected_vendor_id_1 = 987654321

    mc, alias = motion_controller
    vendor_id_0 = mc.configuration.get_vendor_id(alias, axis=0)
    vendor_id_1 = mc.configuration.get_vendor_id(alias, axis=1)

    assert vendor_id_0 == expected_vendor_id_0
    assert vendor_id_1 == expected_vendor_id_1


@pytest.mark.virtual
def test_change_node_id_exception(motion_controller):
    mc, alias = motion_controller
    with pytest.raises(ValueError):
        mc.configuration.change_node_id(32, alias)


@pytest.mark.virtual
@pytest.mark.smoke
def test_set_velocity_pid(motion_controller_teardown):
    mc, alias = motion_controller_teardown
    kp_test = 1
    ki_test = 2
    kd_test = 3
    mc.configuration.set_velocity_pid(kp_test, ki_test, kd_test, servo=alias)
    kp_reg = mc.communication.get_register(VELOCITY_LOOP_KP_REGISTER, servo=alias)
    ki_reg = mc.communication.get_register(VELOCITY_LOOP_KI_REGISTER, servo=alias)
    kd_reg = mc.communication.get_register(VELOCITY_LOOP_KD_REGISTER, servo=alias)
    assert kp_test == kp_reg
    assert ki_test == ki_reg
    assert kd_test == kd_reg


@pytest.mark.virtual
@pytest.mark.smoke
def test_set_position_pid(motion_controller_teardown):
    mc, alias = motion_controller_teardown
    kp_test = 1
    ki_test = 2
    kd_test = 3
    mc.configuration.set_position_pid(kp_test, ki_test, kd_test, servo=alias)
    kp_reg = mc.communication.get_register(POSITION_LOOP_KP_REGISTER, servo=alias)
    ki_reg = mc.communication.get_register(POSITION_LOOP_KI_REGISTER, servo=alias)
    kd_reg = mc.communication.get_register(POSITION_LOOP_KD_REGISTER, servo=alias)
    assert kp_test == kp_reg
    assert ki_test == ki_reg
    assert kd_test == kd_reg


@pytest.mark.virtual
@pytest.mark.smoke
def test_get_set_rated_current(motion_controller):
    mc, alias = motion_controller
    initial_rated_current = mc.communication.get_register(RATED_CURRENT_REGISTER, servo=alias)
    read_rated_current = mc.configuration.get_rated_current(alias)
    assert pytest.approx(initial_rated_current) == read_rated_current
    test_rated_current = 1.23
    mc.configuration.set_rated_current(test_rated_current, servo=alias)
    read_test_rated_current = mc.communication.get_register(RATED_CURRENT_REGISTER, servo=alias)
    assert pytest.approx(test_rated_current) == read_test_rated_current
    # Teardown
    mc.communication.set_register(RATED_CURRENT_REGISTER, initial_rated_current, servo=alias)


@pytest.mark.virtual
@pytest.mark.smoke
def test_get_max_current(motion_controller):
    mc, alias = motion_controller
    real_max_current = mc.communication.get_register(MAX_CURRENT_REGISTER, servo=alias)
    test_max_current = mc.configuration.get_max_current(alias)
    assert pytest.approx(real_max_current) == test_max_current


@pytest.mark.virtual
def test_set_commutation_mode(motion_controller):
    input_value = CommutationMode.SINUSOIDAL
    mc, alias = motion_controller
    mc.configuration.set_commutation_mode(input_value, servo=alias)
    output_value = mc.communication.get_register(COMMUTATION_MODE_REGISTER, servo=alias)
    assert pytest.approx(input_value) == output_value


@pytest.mark.virtual
def test_get_commutation_mode(motion_controller):
    mc, alias = motion_controller
    test_value = mc.configuration.get_commutation_mode(servo=alias)
    reg_value = mc.communication.get_register(COMMUTATION_MODE_REGISTER, servo=alias)
    assert test_value == reg_value


@pytest.mark.virtual
def test_get_bus_voltage(motion_controller):
    mc, alias = motion_controller
    test_value = mc.configuration.get_bus_voltage(servo=alias)
    reg_value = mc.communication.get_register(BUS_VOLTAGE_REGISTER, servo=alias)
    assert pytest.approx(test_value) == reg_value


@pytest.mark.virtual
def test_set_pos_to_vel_ratio(motion_controller):
    input_value = 1.0
    mc, alias = motion_controller
    mc.configuration.set_pos_to_vel_ratio(input_value, servo=alias)
    output_value = mc.communication.get_register(POSITION_TO_VELOCITY_RATIO_REGISTER, servo=alias)
    assert pytest.approx(input_value) == output_value


@pytest.mark.virtual
def test_get_pos_to_vel_ratio(motion_controller):
    mc, alias = motion_controller
    test_value = mc.configuration.get_pos_to_vel_ratio(servo=alias)
    reg_value = mc.communication.get_register(POSITION_TO_VELOCITY_RATIO_REGISTER, servo=alias)
    assert test_value == reg_value


@pytest.mark.virtual
@pytest.mark.parametrize("filter_type", [FilterType.LOWPASS, FilterType.HIGHPASS])
@pytest.mark.parametrize("filter_number", [FilterNumber.FILTER1, FilterNumber.FILTER2])
@pytest.mark.parametrize(
    "filter_signal", [FilterSignal.CURRENT_FEEDBACK, FilterSignal.VELOCITY_REFERENCE]
)
def test_configure_filter(motion_controller, filter_type, filter_number, filter_signal):
    mc, alias = motion_controller
    frequency = 10
    q_factor = 1.3
    gain = 1.2
    mc.configuration.configure_filter(
        filter_signal,
        filter_number,
        filter_type=filter_type,
        frequency=frequency,
        q_factor=q_factor,
        gain=gain,
        servo=alias,
    )

    reg_type = FILTER_TYPE_REGISTER.format(filter_signal.value, filter_number.value)
    read_type = mc.communication.get_register(reg_type, servo=alias)
    assert pytest.approx(read_type) == filter_type

    reg_freq = FILTER_FREQ_REGISTER.format(filter_signal.value, filter_number.value)
    read_freq = mc.communication.get_register(reg_freq, servo=alias)
    assert pytest.approx(read_freq) == frequency

    reg_q_factor = FILTER_Q_REGISTER.format(filter_signal.value, filter_number.value)
    read_q_factor = mc.communication.get_register(reg_q_factor, servo=alias)
    assert pytest.approx(read_q_factor) == q_factor

    reg_gain = FILTER_GAIN_REGISTER.format(filter_signal.value, filter_number.value)
    read_gain = mc.communication.get_register(reg_gain, servo=alias)
    assert pytest.approx(read_gain) == gain


@pytest.mark.virtual
def test_load_configuration_file_not_found(motion_controller):
    file_path = "test_file.xcf"
    mc, alias = motion_controller
    with pytest.raises(FileNotFoundError):
        mc.configuration.load_configuration(file_path, servo=alias)


@pytest.mark.parametrize(
    "function, wrong_value",
    [
        ("get_max_velocity", 1),
        ("get_position_and_velocity_loop_rate", 1.0),
        ("get_current_loop_rate", 1.0),
        ("get_power_stage_frequency", 1.0),
        ("get_status_word", 1.0),
        ("get_phasing_mode", 1.0),
        ("get_motor_pair_poles", 1.0),
        ("get_sto_status", 1.0),
        ("get_rated_current", 1),
        ("get_max_current", 1),
        ("get_commutation_mode", 1.0),
        ("get_bus_voltage", 1),
        ("get_pos_to_vel_ratio", 1),
    ],
)
@pytest.mark.virtual
def test_wrong_type_exception(mocker, motion_controller, function, wrong_value):
    mc, alias = motion_controller
    mocker.patch.object(mc.communication, "get_register", return_value=wrong_value)
    with pytest.raises(TypeError):
        getattr(mc.configuration, function)(servo=alias)


@pytest.mark.virtual
def test_get_phasing_mode_invalid(mocker, motion_controller):
    mc, alias = motion_controller
    invalid_enum_value = 8
    mocker.patch.object(mc.communication, "get_register", return_value=invalid_enum_value)
    phasing_mode = mc.configuration.get_phasing_mode(servo=alias)
    assert phasing_mode == invalid_enum_value


@pytest.mark.virtual
def test_change_tcp_ip_parameters_exception(mocker, motion_controller):
    mc, alias = motion_controller
    mocker.patch.object(mc, "_get_drive", return_value=EthercatServo)
    with pytest.raises(IMException):
        mc.configuration.change_tcp_ip_parameters(
            "192.168.2.22", "255.255.0.0", "192.168.2.1", servo=alias
        )


@pytest.mark.parametrize(
    "function",
    [
        "store_tcp_ip_parameters",
        "restore_tcp_ip_parameters",
    ],
)
@pytest.mark.virtual
def test_store_restore_tcp_ip_parameters_exception(mocker, motion_controller, function):
    mc, alias = motion_controller
    mocker.patch.object(mc, "_get_drive", return_value=EthercatServo)
    with pytest.raises(IMException):
        getattr(mc.configuration, function)(servo=alias)


@pytest.mark.parametrize(
    "subnode, expected_result",
    [
        (0, TYPE_SUBNODES.COCO),
        (1, TYPE_SUBNODES.MOCO),
    ],
)
@pytest.mark.virtual
def test_get_subnode_type(motion_controller, subnode, expected_result):
    mc, alias = motion_controller
    assert mc.configuration.get_subnode_type(subnode) == expected_result


@pytest.mark.virtual
def test_get_subnode_type_exception(motion_controller):
    mc, alias = motion_controller
    with pytest.raises(ValueError):
        mc.configuration.get_subnode_type(-1)



from typing import TYPE_CHECKING, Dict, List, Optional, Tuple, Union

import ingenialogger
from ingenialink.enums.register import REG_DTYPE

from ingeniamotion.enums import MonitoringSoCConfig, MonitoringSoCType, MonitoringVersion
from ingeniamotion.exceptions import IMStatusWordError
from ingeniamotion.metaclass import DEFAULT_SERVO

if TYPE_CHECKING:
    from ingeniamotion.motion_controller import MotionController

from ingeniamotion.monitoring.base_monitoring import Monitoring, check_monitoring_disabled


class MonitoringV1(Monitoring):
    EOC_TRIGGER_NUMBER_SAMPLES = 3

    TRIGGER_CYCLIC_RISING_EDGE = 2
    TRIGGER_CYCLIC_FALLING_EDGE = 4

    MONITORING_NUMBER_TRIGGER_REPETITIONS_REGISTER = "MON_CFG_TRIGGER_REPETITIONS"
    MONITOR_END_CONDITION_TYPE_REGISTER = "MON_CFG_EOC_TYPE"

    EDGE_CONDITION_REGISTER = {
        MonitoringSoCConfig.TRIGGER_CONFIG_RISING: "MON_CFG_RISING_CONDITION",
        MonitoringSoCConfig.TRIGGER_CONFIG_FALLING: "MON_CFG_FALLING_CONDITION",
    }

    def __init__(self, mc: "MotionController", servo: str = DEFAULT_SERVO) -> None:
        super().__init__(mc, servo)
        self._version = mc.capture._check_version(servo)
        self.logger = ingenialogger.get_logger(__name__, drive=mc.servo_name(servo))
        try:
            self.mc.capture.mcb_synchronization(servo=servo)
        except IMStatusWordError:
            self.logger.warning(
                "MCB could not be synchronized. Motor is enabled.", drive=mc.servo_name(servo)
            )

    def __get_old_soc_type(self, edge_condition: MonitoringSoCConfig) -> int:
        if edge_condition == MonitoringSoCConfig.TRIGGER_CONFIG_RISING:
            return self.TRIGGER_CYCLIC_RISING_EDGE
        if edge_condition == MonitoringSoCConfig.TRIGGER_CONFIG_FALLING:
            return self.TRIGGER_CYCLIC_FALLING_EDGE
        raise NotImplementedError("Edge condition is not implementedfor this FW version")

    @check_monitoring_disabled
    def set_trigger(
        self,
        trigger_mode: Union[MonitoringSoCType, int],
        edge_condition: Optional[MonitoringSoCConfig] = None,
        trigger_signal: Optional[Dict[str, str]] = None,
        trigger_value: Union[int, float, None] = None,
    ) -> None:
        self.rearm_monitoring()
        if trigger_mode == MonitoringSoCType.TRIGGER_EVENT_EDGE:
            if trigger_signal is None or trigger_value is None:
                raise TypeError("trigger_signal or trigger_value are None")
            if edge_condition is None:
                raise TypeError("Edge condition is not selected")
            trigger_mode = self.__get_old_soc_type(edge_condition)
            index_reg, level_edge = self._get_reg_index_and_edge_condition_value(
                trigger_signal, trigger_value
            )
            self.__rising_or_falling_edge_trigger(edge_condition, index_reg, level_edge)
        self.mc.communication.set_register(
            self.MONITOR_START_CONDITION_TYPE_REGISTER, trigger_mode, servo=self.servo, axis=0
        )

    def __rising_or_falling_edge_trigger(
        self, edge_condition: MonitoringSoCConfig, index_reg: int, level_edge: int
    ) -> None:
        self.mc.communication.set_register(
            self.MONITORING_INDEX_CHECKER_REGISTER, index_reg, servo=self.servo, axis=0
        )
        self.mc.communication.set_register(
            self.EDGE_CONDITION_REGISTER[edge_condition], level_edge, servo=self.servo, axis=0
        )

    @check_monitoring_disabled
    def configure_number_samples(self, total_num_samples: int, trigger_delay_samples: int) -> None:
        if trigger_delay_samples > total_num_samples:
            raise ValueError("trigger_delay_samples should be less than total_num_samples")
        if trigger_delay_samples < 0:
            raise ValueError("trigger_delay_samples should be a positive number")
        self._check_buffer_size_is_enough(
            total_num_samples, trigger_delay_samples, self.mapped_registers
        )
        if trigger_delay_samples == 0:
            trigger_delay_samples = 1
        if trigger_delay_samples == total_num_samples:
            trigger_delay_samples = total_num_samples - 1
        window_samples = total_num_samples - trigger_delay_samples
        self.mc.communication.set_register(
            self.MONITOR_END_CONDITION_TYPE_REGISTER,
            self.EOC_TRIGGER_NUMBER_SAMPLES,
            servo=self.servo,
            axis=0,
        )
        self.mc.communication.set_register(
            self.MONITORING_TRIGGER_DELAY_SAMPLES_REGISTER,
            trigger_delay_samples,
            servo=self.servo,
            axis=0,
        )
        self.mc.communication.set_register(
            self.MONITORING_WINDOW_NUMBER_SAMPLES_REGISTER, window_samples, servo=self.servo, axis=0
        )
        self.samples_number = total_num_samples
        self.trigger_delay_samples = trigger_delay_samples

    def _check_monitoring_is_ready(self) -> Tuple[bool, Optional[str]]:
        is_enabled = self.mc.capture.is_monitoring_enabled(self.servo)
        result_text = None
        trigger_repetitions = self.mc.communication.get_register(
            self.MONITORING_NUMBER_TRIGGER_REPETITIONS_REGISTER, servo=self.servo, axis=0
        )
        is_ready = is_enabled and trigger_repetitions != 0
        if not is_ready:
            text_is_enabled = "enabled" if is_enabled else "disabled"
            result_text = (
                "Can't read monitoring data because monitoring is not ready."
                " MON_CFG_TRIGGER_REPETITIONS is {}. Monitoring is {}.".format(
                    trigger_repetitions, text_is_enabled
                )
            )
        return is_ready, result_text

    def _check_data_is_ready(self) -> bool:
        monit_nmb_blocks = self.mc.communication.get_register(
            self.MONITORING_ACTUAL_NUMBER_SAMPLES_REGISTER, servo=self.servo, axis=0
        )
        if not isinstance(monit_nmb_blocks, int):
            raise TypeError("Actual number of monitoring samples value has to be an integer")
        data_is_ready = monit_nmb_blocks > 0
        if self._version == MonitoringVersion.MONITORING_V2:
            data_is_ready &= self.mc.capture.is_frame_available(self.servo, version=self._version)
        return data_is_ready

    def rearm_monitoring(self) -> None:
        self.mc.communication.set_register(
            self.MONITORING_NUMBER_TRIGGER_REPETITIONS_REGISTER, 1, servo=self.servo, axis=0
        )

    def _check_buffer_size_is_enough(
        self,
        total_samples: int,
        trigger_delay_samples: int,
        registers: List[Dict[str, Union[int, str, REG_DTYPE]]],
    ) -> None:
        n_sample = max(total_samples - trigger_delay_samples, trigger_delay_samples)
        max_size = self.max_sample_number // 2
        self._check_samples_and_max_size(n_sample, max_size, registers)



from typing import TYPE_CHECKING, Callable, Dict, List, Optional, Tuple, Union

import ingenialogger
from ingenialink.enums.register import REG_DTYPE

from ingeniamotion.enums import (
    MonitoringProcessStage,
    MonitoringSoCConfig,
    MonitoringSoCType,
    MonitoringVersion,
)
from ingeniamotion.metaclass import DEFAULT_SERVO

if TYPE_CHECKING:
    from ingeniamotion.motion_controller import MotionController

from ingeniamotion.monitoring.base_monitoring import Monitoring, check_monitoring_disabled


class MonitoringV3(Monitoring):
    MONITORING_REARM_REGISTER = "MON_REARM"
    MONITOR_START_CONDITION_CONFIG_REGISTER = "MON_CFG_EOC_TYPE"
    MONITORING_TRIGGER_THRESHOLD_REGISTER = "MON_CFG_RISING_CONDITION"

    def __init__(self, mc: "MotionController", servo: str = DEFAULT_SERVO) -> None:
        super().__init__(mc, servo)
        self._version = MonitoringVersion.MONITORING_V3
        self.logger = ingenialogger.get_logger(__name__, drive=mc.servo_name(servo))

    @check_monitoring_disabled
    def set_trigger(
        self,
        trigger_mode: MonitoringSoCType,
        edge_condition: Optional[MonitoringSoCConfig] = None,
        trigger_signal: Optional[Dict[str, str]] = None,
        trigger_value: Union[int, float, None] = None,
    ) -> None:
        self.mc.communication.set_register(
            self.MONITOR_START_CONDITION_TYPE_REGISTER, trigger_mode, servo=self.servo, axis=0
        )
        if trigger_mode == MonitoringSoCType.TRIGGER_EVENT_EDGE:
            if trigger_signal is None or trigger_value is None:
                raise TypeError("trigger_signal or trigger_value are None")
            if edge_condition is None:
                raise TypeError("Edge condition is not selected")
            index_reg, level_edge = self._get_reg_index_and_edge_condition_value(
                trigger_signal, trigger_value
            )
            self.__rising_or_falling_edge_trigger(edge_condition, index_reg, level_edge)

    def __rising_or_falling_edge_trigger(
        self, edge_condition: MonitoringSoCConfig, index_reg: int, level_edge: int
    ) -> None:
        self.mc.communication.set_register(
            self.MONITOR_START_CONDITION_CONFIG_REGISTER, edge_condition, servo=self.servo, axis=0
        )
        self.mc.communication.set_register(
            self.MONITORING_TRIGGER_THRESHOLD_REGISTER, level_edge, servo=self.servo, axis=0
        )
        self.mc.communication.set_register(
            self.MONITORING_INDEX_CHECKER_REGISTER, index_reg, servo=self.servo, axis=0
        )

    @check_monitoring_disabled
    def configure_number_samples(self, total_num_samples: int, trigger_delay_samples: int) -> None:
        if trigger_delay_samples > total_num_samples:
            raise ValueError("trigger_delay_samples should be less than total_num_samples")
        if trigger_delay_samples < 0:
            raise ValueError("trigger_delay_samples should be a positive number")
        self._check_buffer_size_is_enough(
            total_num_samples, trigger_delay_samples, self.mapped_registers
        )

        self.mc.communication.set_register(
            self.MONITORING_TRIGGER_DELAY_SAMPLES_REGISTER,
            trigger_delay_samples,
            servo=self.servo,
            axis=0,
        )
        self.mc.communication.set_register(
            self.MONITORING_WINDOW_NUMBER_SAMPLES_REGISTER,
            total_num_samples,
            servo=self.servo,
            axis=0,
        )
        self.samples_number = total_num_samples
        self.trigger_delay_samples = trigger_delay_samples

    def _check_monitoring_is_ready(self) -> Tuple[bool, Optional[str]]:
        is_enabled = self.mc.capture.is_monitoring_enabled(self.servo)
        result_text = None
        monitoring_stage = self.mc.capture.get_monitoring_process_stage(self.servo, self._version)
        is_ready = is_enabled and monitoring_stage != MonitoringProcessStage.INIT_STAGE
        if not is_ready:
            text_is_enabled = "enabled" if is_enabled else "disabled"
            result_text = (
                "Can't read monitoring data because monitoring is not ready."
                " Monitoring stage is {}. Monitoring is {}.".format(
                    monitoring_stage.name, text_is_enabled
                )
            )
        return is_ready, result_text

    # TODO Study remove progress_callback
    def read_monitoring_data(
        self,
        timeout: Optional[float] = None,
        progress_callback: Optional[Callable[[MonitoringProcessStage, float], None]] = None,
    ) -> List[List[Union[int, float]]]:
        drive = self.mc.servos[self.servo]
        data_array = super().read_monitoring_data(
            timeout=timeout, progress_callback=progress_callback
        )
        drive.monitoring_remove_data()
        return data_array

    def _check_data_is_ready(self) -> bool:
        monit_nmb_blocks = self.mc.communication.get_register(
            self.MONITORING_ACTUAL_NUMBER_SAMPLES_REGISTER, servo=self.servo, axis=0
        )
        if not isinstance(monit_nmb_blocks, int):
            raise TypeError("Actual number of monitoring samples value has to be an integer")
        data_is_ready = monit_nmb_blocks > 0
        data_is_ready &= self.mc.capture.is_frame_available(self.servo, version=self._version)
        return data_is_ready

    def rearm_monitoring(self) -> None:
        self.mc.communication.set_register(
            self.MONITORING_REARM_REGISTER, 1, servo=self.servo, axis=0
        )

    def _check_buffer_size_is_enough(
        self,
        total_samples: int,
        trigger_delay_samples: int,
        registers: List[Dict[str, Union[int, str, REG_DTYPE]]],
    ) -> None:
        n_sample = total_samples
        max_size = self.max_sample_number
        self._check_samples_and_max_size(n_sample, max_size, registers)
